from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class RealEsrGan(ReplicateNode):
    """Real-ESRGAN with optional face correction and adjustable upscale"""

    @classmethod
    def replicate_model_id(cls):
        return "nightmareai/real-esrgan:350d32041630ffbe63c8352783a26d94126809164e54085352f8326e53999085"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/90563818-e1bc-4585-a626-62b5ab5ff22d/ApplicationFrameHost_3Bs5DdjGnX.png",
            "created_at": "2022-06-22T01:08:15.680509Z",
            "description": "Real-ESRGAN with optional face correction and adjustable upscale",
            "github_url": "https://github.com/NightmareAI/Real-ESRGAN",
            "license_url": "https://github.com/NightmareAI/Real-ESRGAN/blob/master/LICENSE",
            "name": "real-esrgan",
            "owner": "nightmareai",
            "paper_url": None,
            "run_count": 42537712,
            "url": "https://replicate.com/nightmareai/real-esrgan",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    scale: float = Field(
        title="Scale",
        description="Factor to scale image by",
        ge=0.0,
        le=10.0,
        default=4,
    )
    face_enhance: bool = Field(
        title="Face Enhance", description="Face enhance", default=True
    )


class Sd_model(str, Enum):
    EPICREALISM_NATURALSINRC1VAE_SAFETENSORS__84D76A0328 = (
        "epicrealism_naturalSinRC1VAE.safetensors [84d76a0328]"
    )
    JUGGERNAUT_REBORN_SAFETENSORS__338B85BC4F = (
        "juggernaut_reborn.safetensors [338b85bc4f]"
    )
    FLAT2DANIMERGE_V45SHARP_SAFETENSORS = "flat2DAnimerge_v45Sharp.safetensors"


class Scheduler(str, Enum):
    DPM___2M_KARRAS = "DPM++ 2M Karras"
    DPM___SDE_KARRAS = "DPM++ SDE Karras"
    DPM___2M_SDE_EXPONENTIAL = "DPM++ 2M SDE Exponential"
    DPM___2M_SDE_KARRAS = "DPM++ 2M SDE Karras"
    EULER_A = "Euler a"
    EULER = "Euler"
    LMS = "LMS"
    HEUN = "Heun"
    DPM2 = "DPM2"
    DPM2_A = "DPM2 a"
    DPM___2S_A = "DPM++ 2S a"
    DPM___2M = "DPM++ 2M"
    DPM___SDE = "DPM++ SDE"
    DPM___2M_SDE = "DPM++ 2M SDE"
    DPM___2M_SDE_HEUN = "DPM++ 2M SDE Heun"
    DPM___2M_SDE_HEUN_KARRAS = "DPM++ 2M SDE Heun Karras"
    DPM___2M_SDE_HEUN_EXPONENTIAL = "DPM++ 2M SDE Heun Exponential"
    DPM___3M_SDE = "DPM++ 3M SDE"
    DPM___3M_SDE_KARRAS = "DPM++ 3M SDE Karras"
    DPM___3M_SDE_EXPONENTIAL = "DPM++ 3M SDE Exponential"
    DPM_FAST = "DPM fast"
    DPM_ADAPTIVE = "DPM adaptive"
    LMS_KARRAS = "LMS Karras"
    DPM2_KARRAS = "DPM2 Karras"
    DPM2_A_KARRAS = "DPM2 a Karras"
    DPM___2S_A_KARRAS = "DPM++ 2S a Karras"
    RESTART = "Restart"
    DDIM = "DDIM"
    PLMS = "PLMS"
    UNIPC = "UniPC"


class Tiling_width(int, Enum):
    _16 = 16
    _32 = 32
    _48 = 48
    _64 = 64
    _80 = 80
    _96 = 96
    _112 = 112
    _128 = 128
    _144 = 144
    _160 = 160
    _176 = 176
    _192 = 192
    _208 = 208
    _224 = 224
    _240 = 240
    _256 = 256


class Tiling_height(int, Enum):
    _16 = 16
    _32 = 32
    _48 = 48
    _64 = 64
    _80 = 80
    _96 = 96
    _112 = 112
    _128 = 128
    _144 = 144
    _160 = 160
    _176 = 176
    _192 = 192
    _208 = 208
    _224 = 224
    _240 = 240
    _256 = 256


class ClarityUpscaler(ReplicateNode):
    """High resolution image Upscaler and Enhancer. Use at ClarityAI.cc. A free Magnific alternative. Twitter/X: @philz1337x"""

    @classmethod
    def replicate_model_id(cls):
        return "philz1337x/clarity-upscaler:b8a46b09384dc1ac996596bc14058e2b7604971128ee7de709a40d4bbf982d2c"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (40GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/76fbe6df-517d-480a-a6e4-dce383b40bbb/Bildschirmfoto_2024-03-30_um_09.2.png",
            "created_at": "2024-03-15T02:35:32.167345Z",
            "description": "High resolution image Upscaler and Enhancer. Use at ClarityAI.cc. A free Magnific alternative. Twitter/X: @philz1337x",
            "github_url": "https://github.com/philz1337x/clarity-upscaler",
            "license_url": "https://github.com/philz1337x/clarity-upscaler/blob/main/LICENSE.txt",
            "name": "clarity-upscaler",
            "owner": "philz1337x",
            "paper_url": "https://clarityai.cc",
            "run_count": 873564,
            "url": "https://replicate.com/philz1337x/clarity-upscaler",
            "visibility": "public",
            "hardware": "Nvidia A100 (40GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    seed: int = Field(
        title="Seed",
        description="Random seed. Leave blank to randomize the seed",
        default=1337,
    )
    image: ImageRef = Field(default=ImageRef(), description="input image")
    prompt: str = Field(
        title="Prompt",
        description="Prompt",
        default="masterpiece, best quality, highres, <lora:more_details:0.5> <lora:SDXLrender_v2.0:1>",
    )
    dynamic: float = Field(
        title="Dynamic", description="HDR, try from 3 - 9", ge=1.0, le=50.0, default=6
    )
    sd_model: Sd_model = Field(
        description="Stable Diffusion model checkpoint",
        default="juggernaut_reborn.safetensors [338b85bc4f]",
    )
    scheduler: Scheduler = Field(description="scheduler", default="DPM++ 3M SDE Karras")
    creativity: float = Field(
        title="Creativity",
        description="Creativity, try from 0.3 - 0.9",
        ge=0.0,
        le=1.0,
        default=0.35,
    )
    lora_links: str = Field(
        title="Lora Links",
        description="Link to a lora file you want to use in your upscaling. Multiple links possible, seperated by comma",
        default="",
    )
    downscaling: bool = Field(
        title="Downscaling",
        description="Downscale the image before upscaling. Can improve quality and speed for images with high resolution but lower quality",
        default=False,
    )
    resemblance: float = Field(
        title="Resemblance",
        description="Resemblance, try from 0.3 - 1.6",
        ge=0.0,
        le=3.0,
        default=0.6,
    )
    scale_factor: float = Field(
        title="Scale Factor", description="Scale factor", default=2
    )
    tiling_width: Tiling_width = Field(
        description="Fractality, set lower tile width for a high Fractality",
        default=112,
    )
    tiling_height: Tiling_height = Field(
        description="Fractality, set lower tile height for a high Fractality",
        default=144,
    )
    custom_sd_model: str = Field(title="Custom Sd Model", default="")
    negative_prompt: str = Field(
        title="Negative Prompt",
        description="Negative Prompt",
        default="(worst quality, low quality, normal quality:2) JuggernautNegative-neg",
    )
    num_inference_steps: int = Field(
        title="Num Inference Steps",
        description="Number of denoising steps",
        ge=1.0,
        le=100.0,
        default=18,
    )
    downscaling_resolution: int = Field(
        title="Downscaling Resolution",
        description="Downscaling resolution",
        default=768,
    )


class Resolution(str, Enum):
    ORIGINAL = "original"
    _1024 = "1024"
    _2048 = "2048"


class MagicImageRefiner(ReplicateNode):
    """A better alternative to SDXL refiners, providing a lot of quality and detail. Can also be used for inpainting or upscaling."""

    @classmethod
    def replicate_model_id(cls):
        return "batouresearch/magic-image-refiner:507ddf6f977a7e30e46c0daefd30de7d563c72322f9e4cf7cbac52ef0f667b13"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/H3ZmqoAgsBonKFilPafiEsvYsc2FnjD8EW3vMt6KpkYfd0ISA/out-0.png",
            "created_at": "2024-01-03T16:55:24.594128Z",
            "description": "A better alternative to SDXL refiners, providing a lot of quality and detail. Can also be used for inpainting or upscaling.",
            "github_url": "https://github.com/BatouResearch/magic-image-refiner",
            "license_url": None,
            "name": "magic-image-refiner",
            "owner": "batouresearch",
            "paper_url": None,
            "run_count": 517166,
            "url": "https://replicate.com/batouresearch/magic-image-refiner",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    hdr: float = Field(
        title="Hdr",
        description="HDR improvement over the original image",
        ge=0.0,
        le=1.0,
        default=0,
    )
    mask: ImageRef = Field(
        default=ImageRef(),
        description="When provided, refines some section of the image. Must be the same size as the image",
    )
    seed: int | None = Field(title="Seed", description="Seed", default=None)
    image: ImageRef = Field(default=ImageRef(), description="Image to refine")
    steps: int = Field(title="Steps", description="Steps", default=20)
    prompt: str | None = Field(
        title="Prompt", description="Prompt for the model", default=None
    )
    scheduler: Scheduler = Field(description="Choose a scheduler.", default="DDIM")
    creativity: float = Field(
        title="Creativity",
        description="Denoising strength. 1 means total destruction of the original image",
        ge=0.0,
        le=1.0,
        default=0.25,
    )
    guess_mode: bool = Field(
        title="Guess Mode",
        description="In this mode, the ControlNet encoder will try best to recognize the content of the input image even if you remove all prompts. The `guidance_scale` between 3.0 and 5.0 is recommended.",
        default=False,
    )
    resolution: Resolution = Field(description="Image resolution", default="original")
    resemblance: float = Field(
        title="Resemblance",
        description="Conditioning scale for controlnet",
        ge=0.0,
        le=1.0,
        default=0.75,
    )
    guidance_scale: float = Field(
        title="Guidance Scale",
        description="Scale for classifier-free guidance",
        ge=0.1,
        le=30.0,
        default=7,
    )
    negative_prompt: str = Field(
        title="Negative Prompt",
        description="Negative prompt",
        default="teeth, tooth, open mouth, longbody, lowres, bad anatomy, bad hands, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, mutant",
    )


class Scale(int, Enum):
    _2 = 2
    _4 = 4
    _8 = 8


class ruDallE_SR(ReplicateNode):
    """Real-ESRGAN super-resolution model from ruDALL-E"""

    @classmethod
    def replicate_model_id(cls):
        return "cjwbw/rudalle-sr:32fdb2231d00a10d33754cc2ba794a2dfec94216579770785849ce6f149dbc69"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/mgxm/588170af-559f-454e-967d-8fb6c7f8304b/out.png",
            "created_at": "2021-11-04T18:36:03.485750Z",
            "description": "Real-ESRGAN super-resolution model from ruDALL-E",
            "github_url": "https://github.com/CJWBW/rudalle-sr",
            "license_url": "https://github.com/chenxwh/rudalle-sr/blob/main/LICENSE.txt",
            "name": "rudalle-sr",
            "owner": "cjwbw",
            "paper_url": "https://arxiv.org/abs/2107.10833",
            "run_count": 464731,
            "url": "https://replicate.com/cjwbw/rudalle-sr",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    scale: Scale = Field(description="Choose up-scaling factor", default=4)


class HighResolutionControlNetTile(ReplicateNode):
    """Fermat.app open-source implementation of an efficient ControlNet 1.1 tile for high-quality upscales. Increase the creativity to encourage hallucination."""

    @classmethod
    def replicate_model_id(cls):
        return "batouresearch/high-resolution-controlnet-tile:4af11083a13ebb9bf97a88d7906ef21cf79d1f2e5fa9d87b70739ce6b8113d29"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (80GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/pbxt/8Reu1zg6zAV6GShu5yiyP0ueFch9EddJPlVfIxyiMLWBlwJkA/out-0.png",
            "created_at": "2023-12-08T02:32:38.082772Z",
            "description": "Fermat.app open-source implementation of an efficient ControlNet 1.1 tile for high-quality upscales. Increase the creativity to encourage hallucination.",
            "github_url": "https://github.com/BatouResearch/controlnet-tile-upscale",
            "license_url": None,
            "name": "high-resolution-controlnet-tile",
            "owner": "batouresearch",
            "paper_url": None,
            "run_count": 298038,
            "url": "https://replicate.com/batouresearch/high-resolution-controlnet-tile",
            "visibility": "public",
            "hardware": "Nvidia A100 (80GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    hdr: float = Field(
        title="Hdr",
        description="HDR improvement over the original image",
        ge=0.0,
        le=1.0,
        default=0,
    )
    seed: int | None = Field(title="Seed", description="Seed", default=None)
    image: ImageRef = Field(
        default=ImageRef(), description="Control image for scribble controlnet"
    )
    steps: int = Field(title="Steps", description="Steps", default=20)
    prompt: str | None = Field(
        title="Prompt", description="Prompt for the model", default=None
    )
    scheduler: Scheduler = Field(description="Choose a scheduler.", default="DDIM")
    creativity: float = Field(
        title="Creativity",
        description="Denoising strength. 1 means total destruction of the original image",
        ge=0.0,
        le=1.0,
        default=0.5,
    )
    guess_mode: bool = Field(
        title="Guess Mode",
        description="In this mode, the ControlNet encoder will try best to recognize the content of the input image even if you remove all prompts. The `guidance_scale` between 3.0 and 5.0 is recommended.",
        default=False,
    )
    resolution: Resolution = Field(description="Image resolution", default=2048)
    resemblance: float = Field(
        title="Resemblance",
        description="Conditioning scale for controlnet",
        ge=0.0,
        le=1.0,
        default=0.5,
    )
    guidance_scale: float = Field(
        title="Guidance Scale",
        description="Scale for classifier-free guidance",
        ge=0.1,
        le=30.0,
        default=7,
    )
    negative_prompt: str = Field(
        title="Negative Prompt",
        description="Negative prompt",
        default="teeth, tooth, open mouth, longbody, lowres, bad anatomy, bad hands, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, mutant",
    )


class Upscaler(str, Enum):
    _4X_NMKD_SIAX_200K = "4x_NMKD-Siax_200k"
    _4X_ULTRASHARP = "4x-UltraSharp"
    REALESRGAN_X4PLUS = "RealESRGAN_x4plus"
    REALESRGAN_X4PLUS_ANIME_6B = "RealESRGAN_x4plus_anime_6B"


class Mode_type(str, Enum):
    LINEAR = "Linear"
    CHESS = "Chess"
    NONE = "None"


class Sampler_name(str, Enum):
    EULER = "euler"
    EULER_ANCESTRAL = "euler_ancestral"
    HEUN = "heun"
    DPM_2 = "dpm_2"
    DPM_2_ANCESTRAL = "dpm_2_ancestral"
    LMS = "lms"
    DPM_FAST = "dpm_fast"
    DPM_ADAPTIVE = "dpm_adaptive"
    DPMPP_2S_ANCESTRAL = "dpmpp_2s_ancestral"
    DPMPP_SDE = "dpmpp_sde"
    DPMPP_SDE_GPU = "dpmpp_sde_gpu"
    DPMPP_2M = "dpmpp_2m"
    DPMPP_2M_SDE = "dpmpp_2m_sde"
    DPMPP_2M_SDE_GPU = "dpmpp_2m_sde_gpu"
    DPMPP_3M_SDE = "dpmpp_3m_sde"
    DPMPP_3M_SDE_GPU = "dpmpp_3m_sde_gpu"
    DPMPP = "dpmpp"
    DDIM = "ddim"
    UNI_PC = "uni_pc"
    UNI_PC_BH2 = "uni_pc_bh2"


class Seam_fix_mode(str, Enum):
    NONE = "None"
    BAND_PASS = "Band Pass"
    HALF_TILE = "Half Tile"
    HALF_TILE___INTERSECTIONS = "Half Tile + Intersections"


class UltimateSDUpscale(ReplicateNode):
    """Ultimate SD Upscale with ControlNet Tile"""

    @classmethod
    def replicate_model_id(cls):
        return "fewjative/ultimate-sd-upscale:5daf1012d946160622cd1bd45ed8f12d9675d24659276ccfe24804035f3b3ad7"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (40GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/ec52df9a-be01-4c74-af9d-a14df03bf03a/output.png",
            "created_at": "2023-11-14T08:41:40.364739Z",
            "description": "Ultimate SD Upscale with ControlNet Tile",
            "github_url": "https://github.com/fewjative/cog-ultimate-sd-upscale",
            "license_url": None,
            "name": "ultimate-sd-upscale",
            "owner": "fewjative",
            "paper_url": None,
            "run_count": 93784,
            "url": "https://replicate.com/fewjative/ultimate-sd-upscale",
            "visibility": "public",
            "hardware": "Nvidia A100 (40GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    cfg: float = Field(title="Cfg", description="CFG", default=8)
    seed: int | None = Field(
        title="Seed", description="Sampling seed, leave Empty for Random", default=None
    )
    image: ImageRef = Field(default=ImageRef(), description="Input image")
    steps: int = Field(title="Steps", description="Steps", default=20)
    denoise: float = Field(title="Denoise", description="Denoise", default=0.2)
    upscaler: Upscaler = Field(description="Upscaler", default="4x-UltraSharp")
    mask_blur: int = Field(title="Mask Blur", description="Mask Blur", default=8)
    mode_type: Mode_type = Field(description="Mode Type", default="Linear")
    scheduler: Scheduler = Field(description="Scheduler", default="normal")
    tile_width: int = Field(title="Tile Width", description="Tile Width", default=512)
    upscale_by: float = Field(title="Upscale By", description="Upscale By", default=2)
    tile_height: int = Field(
        title="Tile Height", description="Tile Height", default=512
    )
    sampler_name: Sampler_name = Field(description="Sampler", default="euler")
    tile_padding: int = Field(
        title="Tile Padding", description="Tile Padding", default=32
    )
    seam_fix_mode: Seam_fix_mode = Field(description="Seam Fix Mode", default="None")
    seam_fix_width: int = Field(
        title="Seam Fix Width", description="Seam Fix Width", default=64
    )
    negative_prompt: str = Field(
        title="Negative Prompt", description="Negative Prompt", default=""
    )
    positive_prompt: str = Field(
        title="Positive Prompt",
        description="Positive Prompt",
        default="Hey! Have a nice day :D",
    )
    seam_fix_denoise: float = Field(
        title="Seam Fix Denoise", description="Seam Fix Denoise", default=1
    )
    seam_fix_padding: int = Field(
        title="Seam Fix Padding", description="Seam Fix Padding", default=16
    )
    seam_fix_mask_blur: int = Field(
        title="Seam Fix Mask Blur", description="Seam Fix Mask Blur", default=8
    )
    controlnet_strength: float = Field(
        title="Controlnet Strength", description="ControlNet Strength", default=1
    )
    force_uniform_tiles: bool = Field(
        title="Force Uniform Tiles", description="Force Uniform Tiles", default=True
    )
    use_controlnet_tile: bool = Field(
        title="Use Controlnet Tile", description="Use ControlNet Tile", default=True
    )


class Noise(int, Enum):
    _15 = 15
    _25 = 25
    _50 = 50


class Task_type(str, Enum):
    REAL_WORLD_IMAGE_SUPER_RESOLUTION_LARGE = "Real-World Image Super-Resolution-Large"
    REAL_WORLD_IMAGE_SUPER_RESOLUTION_MEDIUM = (
        "Real-World Image Super-Resolution-Medium"
    )
    GRAYSCALE_IMAGE_DENOISING = "Grayscale Image Denoising"
    COLOR_IMAGE_DENOISING = "Color Image Denoising"
    JPEG_COMPRESSION_ARTIFACT_REDUCTION = "JPEG Compression Artifact Reduction"


class SwinIR(ReplicateNode):
    """Image Restoration Using Swin Transformer"""

    @classmethod
    def replicate_model_id(cls):
        return "jingyunliang/swinir:660d922d33153019e8c263a3bba265de882e7f4f70396546b6c9c8f9d47a021a"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A100 (40GB) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/mgxm/1e3c0b87-01a7-4795-abac-aaf17479cf84/out.png",
            "created_at": "2021-09-13T19:58:55.156216Z",
            "description": "Image Restoration Using Swin Transformer",
            "github_url": "https://github.com/JingyunLiang/SwinIR",
            "license_url": "https://github.com/JingyunLiang/SwinIR/blob/main/LICENSE",
            "name": "swinir",
            "owner": "jingyunliang",
            "paper_url": "https://arxiv.org/abs/2108.10257",
            "run_count": 5597357,
            "url": "https://replicate.com/jingyunliang/swinir",
            "visibility": "public",
            "hardware": "Nvidia A100 (40GB) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    jpeg: int = Field(
        title="Jpeg",
        description="scale factor, activated for JPEG Compression Artifact Reduction. Leave it as default or arbitrary if other tasks are selected",
        default=40,
    )
    image: ImageRef = Field(default=ImageRef(), description="input image")
    noise: Noise = Field(
        description="noise level, activated for Grayscale Image Denoising and Color Image Denoising. Leave it as default or arbitrary if other tasks are selected",
        default=15,
    )
    task_type: Task_type = Field(
        description="Choose a task", default="Real-World Image Super-Resolution-Large"
    )


class Task(str, Enum):
    CLASSICAL_SR = "classical_sr"
    REAL_SR = "real_sr"
    COMPRESSED_SR = "compressed_sr"


class Swin2SR(ReplicateNode):
    """3 Million Runs! AI Photorealistic Image Super-Resolution and Restoration"""

    @classmethod
    def replicate_model_id(cls):
        return "mv-lab/swin2sr:a01b0512004918ca55d02e554914a9eca63909fa83a29ff0f115c78a7045574f"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/aabde67b-bf5c-4fc8-a4bd-8b2dcba60be6/swin2sr-cover3.png",
            "created_at": "2022-10-28T22:59:05.692845Z",
            "description": "3 Million Runs! AI Photorealistic Image Super-Resolution and Restoration",
            "github_url": "https://github.com/mv-lab/swin2sr",
            "license_url": "https://github.com/mv-lab/swin2sr/blob/main/LICENSE",
            "name": "swin2sr",
            "owner": "mv-lab",
            "paper_url": "https://arxiv.org/abs/2209.11345",
            "run_count": 3480904,
            "url": "https://replicate.com/mv-lab/swin2sr",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    task: Task = Field(description="Choose a task", default="real_sr")
    image: ImageRef = Field(default=ImageRef(), description="Input image")



#https://github.com/SkyTNT/anime-segmentation/tree/main
#Only adapt isnet_is (https://huggingface.co/skytnt/anime-seg/blob/main/isnetis.ckpt)
import torch.nn as nn
import torch
from .isnet import ISNetDIS
import numpy as np
import cv2
from comfy.model_management import get_torch_device
DEVICE = get_torch_device()

class AnimeSegmentation:
    def __init__(self, ckpt_path):
        super(AnimeSegmentation).__init__()
        sd = torch.load(ckpt_path, map_location="cpu")
        self.net = ISNetDIS()
        #gt_encoder isn't used during inference
        self.net.load_state_dict({k.replace("net.", ''):v for k, v in sd.items() if k.startswith("net.")})
        self.net = self.net.to(DEVICE)
        self.net.eval()
    
    def get_mask(self, input_img, s=640):
        input_img = (input_img / 255).astype(np.float32)
        if s == 0:
            img_input = np.transpose(input_img, (2, 0, 1))
            img_input = img_input[np.newaxis, :]
            tmpImg = torch.from_numpy(img_input).float().to(DEVICE)
            with torch.no_grad():
                pred = self.net(tmpImg)[0][0].sigmoid() #https://github.com/SkyTNT/anime-segmentation/blob/main/train.py#L92C20-L92C47
                pred = pred.cpu().numpy()[0]
                pred = np.transpose(pred, (1, 2, 0))
                #pred = pred[:, :, np.newaxis]
                return pred

        h, w = h0, w0 = input_img.shape[:-1]
        h, w = (s, int(s * w / h)) if h > w else (int(s * h / w), s)
        ph, pw = s - h, s - w
        img_input = np.zeros([s, s, 3], dtype=np.float32)
        img_input[ph // 2:ph // 2 + h, pw // 2:pw // 2 + w] = cv2.resize(input_img, (w, h))
        img_input = np.transpose(img_input, (2, 0, 1))
        img_input = img_input[np.newaxis, :]
        tmpImg = torch.from_numpy(img_input).float().to(DEVICE)
        with torch.no_grad():
            pred = self.net(tmpImg)[0][0].sigmoid() #https://github.com/SkyTNT/anime-segmentation/blob/main/train.py#L92C20-L92C47
            pred = pred.cpu().numpy()[0]
            pred = np.transpose(pred, (1, 2, 0))
            pred = pred[ph // 2:ph // 2 + h, pw // 2:pw // 2 + w]
            #pred = cv2.resize(pred, (w0, h0))[:, :, np.newaxis]
            pred = cv2.resize(pred, (w0, h0))
            return pred

    def __call__(self, np_img, img_size):
        mask = self.get_mask(np_img, int(img_size))
        np_img = (mask * np_img + 255 * (1 - mask)).astype(np.uint8)
        mask = (mask * 255).astype(np.uint8)
        #np_img = np.concatenate([np_img, mask], axis=2, dtype=np.uint8)
        #mask = mask.repeat(3, axis=2)
        return mask, np_img




from typing import Optional
from datetime import datetime
from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid


class Job(DBModel):
    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_jobs",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {"id": "S", "user_id": "S", "workflow_id": "S"},
            "global_secondary_indexes": {
                "nodetool_job_user_index": {"user_id": "HASH"},
                "nodetool_job_workflow_index": {"workflow_id": "HASH"},
            },
        }

    id: str = DBField(hash_key=True)
    user_id: str = DBField(default="")
    job_type: str = DBField(default="")
    status: str = DBField(default="starting")
    workflow_id: str = DBField(default="")
    started_at: datetime = DBField(default_factory=datetime.now)
    finished_at: datetime | None = DBField(default=None)
    graph: dict = DBField(default_factory=dict)
    error: str | None = DBField(default=None)
    cost: float | None = DBField(default=None)

    @classmethod
    def find(cls, user_id: str, job_id: str):
        job = cls.get(job_id)
        return job if job and job.user_id == user_id else None

    @classmethod
    def create(cls, workflow_id: str, user_id: str, **kwargs):
        return super().create(
            id=create_time_ordered_uuid(),
            workflow_id=workflow_id,
            user_id=user_id,
            **kwargs
        )

    @classmethod
    def paginate(
        cls,
        user_id: str,
        workflow_id: Optional[str] = None,
        limit: int = 10,
        start_key: Optional[str] = None,
    ):
        if workflow_id:
            return cls.query(
                condition="workflow_id = :workflow_id",
                values={":workflow_id": workflow_id},
                index="nodetool_job_workflow_index",
                limit=limit,
                start_key=start_key,
            )
        elif user_id:
            return cls.query(
                condition="user_id = :user_id",
                values={":user_id": user_id},
                index="nodetool_job_user_index",
                limit=limit,
                start_key=start_key,
            )
        else:
            raise ValueError("Must provide either user_id or workflow_id")



import os
import types
import warnings

import cv2
import numpy as np
import torch
import torchvision.transforms as transforms
from einops import rearrange
from PIL import Image

from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
from .nets.NNET import NNET


# load model
def load_checkpoint(fpath, model):
    ckpt = torch.load(fpath, map_location='cpu')['model']

    load_dict = {}
    for k, v in ckpt.items():
        if k.startswith('module.'):
            k_ = k.replace('module.', '')
            load_dict[k_] = v
        else:
            load_dict[k] = v

    model.load_state_dict(load_dict)
    return model

class NormalBaeDetector:
    def __init__(self, model):
        self.model = model
        self.norm = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="scannet.pt"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        args = types.SimpleNamespace()
        args.mode = 'client'
        args.architecture = 'BN'
        args.pretrained = 'scannet'
        args.sampling_ratio = 0.4
        args.importance_ratio = 0.7
        model = NNET(args)
        model = load_checkpoint(model_path, model)
        model.eval()

        return cls(model)

    def to(self, device):
        self.model.to(device)
        return self


    def __call__(self, input_image, detect_resolution=512, output_type="pil", upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        device = next(iter(self.model.parameters())).device
        image_normal = detected_map
        with torch.no_grad():
            image_normal = torch.from_numpy(image_normal).float().to(device)
            image_normal = image_normal / 255.0
            image_normal = rearrange(image_normal, 'h w c -> 1 c h w')
            image_normal = self.norm(image_normal)

            normal = self.model(image_normal)
            normal = normal[0][-1][:, :3]
            # d = torch.sum(normal ** 2.0, dim=1, keepdim=True) ** 0.5
            # d = torch.maximum(d, torch.ones_like(d) * 1e-5)
            # normal /= d
            normal = ((normal + 1) * 0.5).clip(0, 1)

            normal = rearrange(normal[0], 'c h w -> h w c').cpu().numpy()
            normal_image = (normal * 255.0).clip(0, 255).astype(np.uint8)

        detected_map = remove_pad(HWC3(normal_image))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map
    


from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management
import numpy as np

class MIDAS_Normal_Map_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            a =  ("FLOAT", {"default": np.pi * 2.0, "min": 0.0, "max": np.pi * 5.0, "step": 0.05}),
            bg_threshold = ("FLOAT", {"default": 0.1, "min": 0, "max": 1, "step": 0.05})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Normal and Depth Estimators"

    def execute(self, image, a, bg_threshold, resolution=512, **kwargs):
        from controlnet_aux.midas import MidasDetector

        model = MidasDetector.from_pretrained().to(model_management.get_torch_device())
        #Dirty hack :))
        cb = lambda image, **kargs: model(image, **kargs)[1]
        out = common_annotator_call(cb, image, resolution=resolution, a=a, bg_th=bg_threshold, depth_and_normal=True)
        del model
        return (out, )

class MIDAS_Depth_Map_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            a =  ("FLOAT", {"default": np.pi * 2.0, "min": 0.0, "max": np.pi * 5.0, "step": 0.05}),
            bg_threshold = ("FLOAT", {"default": 0.1, "min": 0, "max": 1, "step": 0.05})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Normal and Depth Estimators"

    def execute(self, image, a, bg_threshold, resolution=512, **kwargs):
        from controlnet_aux.midas import MidasDetector

        # Ref: https://github.com/lllyasviel/ControlNet/blob/main/gradio_depth2image.py
        model = MidasDetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution, a=a, bg_th=bg_threshold)
        del model
        return (out, )

NODE_CLASS_MAPPINGS = {
    "MiDaS-NormalMapPreprocessor": MIDAS_Normal_Map_Preprocessor,
    "MiDaS-DepthMapPreprocessor": MIDAS_Depth_Map_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "MiDaS-NormalMapPreprocessor": "MiDaS Normal Map",
    "MiDaS-DepthMapPreprocessor": "MiDaS Depth Map"
}


import numpy as np
from nodetool.metadata.types import ImageRef, Tensor
from nodetool.metadata.types import TextRef
from nodetool.workflows.base_node import BaseNode
from nodetool.metadata.types import GPTModel
from nodetool.workflows.processing_context import ProcessingContext

from openai.types.chat import ChatCompletion, ChatCompletionMessageParam
from openai.types.create_embedding_response import CreateEmbeddingResponse
from pydantic import Field

from enum import Enum


class EmbeddingModel(str, Enum):
    TEXT_EMBEDDING_3_LARGE = "text-embedding-3-large"
    TEXT_EMBEDDING_3_SMALL = "text-embedding-3-small"


class ResponseFormat(str, Enum):
    JSON_OBJECT = "json_object"
    TEXT = "text"


class Embedding(BaseNode):
    """
    Generates a vector representation of text for measuring relatedness.
    text, analyse, transform, embeddings, relatedness, search, classification, clustering, recommendations
    Outputs a text embedding vector that quantifies the semantic similarity of the input text to other text strings. An embedding is a vector (list) of floating point numbers. The distance between two vectors measures their relatedness. Small distances suggest high relatedness and large distances suggest low relatedness. Use cases: Search, Clustering, Recommendations, Anomaly detection, Diversity measurement, Classification
    """

    input: str | TextRef = Field(title="Input", default="")
    model: EmbeddingModel = Field(
        title="Model", default=EmbeddingModel.TEXT_EMBEDDING_3_SMALL
    )
    chunk_size: int = 4096

    async def process(self, context: ProcessingContext) -> Tensor:
        input = await context.to_str(self.input)
        # chunk the input into smaller pieces
        chunks = [
            input[i : i + self.chunk_size]
            for i in range(0, len(input), self.chunk_size)
        ]

        response = await context.run_prediction(
            self.id, provider="openai", params={"input": chunks}, model=self.model.value
        )

        res = CreateEmbeddingResponse(**response.json())

        all = [i.embedding for i in res.data]
        avg = np.mean(all, axis=0)
        return Tensor.from_numpy(avg)


class GPT(BaseNode):
    """
    Use GPT models for generating natural language responses based on input prompts.
    text, llm, t2t, ttt, text-to-text, generate, gpt, chat, chatgpt
    Produces natural language text as a response to the input query, leveraging the capabilities of GPT models for various applications.
    """

    model: GPTModel = Field(title="Model", default=GPTModel.GPT3)
    system: str = Field(title="System", default="You are a friendly assistant.")
    prompt: str = Field(title="Prompt", default="")
    image: ImageRef = Field(title="Image", default=ImageRef())
    presence_penalty: float = Field(
        title="Presence Penalty", default=0.0, ge=(-2.0), le=2.0
    )
    frequency_penalty: float = Field(
        title="Frequency Penalty", default=0.0, ge=(-2.0), le=2.0
    )
    temperature: float = Field(title="Temperature", default=1.0, ge=0.0, le=2.0)
    max_tokens: int = Field(title="Max Tokens", default=100, ge=1, le=2048)
    top_p: float = Field(title="Top P", default=1.0, ge=0.0, le=1.0)
    response_format: ResponseFormat = Field(
        title="Response Format", default=ResponseFormat.TEXT
    )

    async def process(self, context: ProcessingContext) -> str:
        content = []
        if self.image.uri != "":
            base64_image = await context.image_to_base64(self.image)
            content.append(
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"},
                }
            )

        content.append(
            {
                "type": "text",
                "text": self.prompt,
            }
        )

        messages: list[ChatCompletionMessageParam] = [
            {"role": "system", "content": self.system},
            {"role": "user", "content": content},
        ]

        response = await context.run_prediction(
            node_id=self.id,
            provider="openai",
            params={
                "messages": messages,
                "max_tokens": self.max_tokens,
                "temperature": self.temperature,
                "top_p": self.top_p,
                "presence_penalty": self.presence_penalty,
                "frequency_penalty": self.frequency_penalty,
                "response_format": self.response_format,
            },
            model=self.model.value,
        )

        res = ChatCompletion(**response.json())

        assert len(res.choices) > 0
        assert res.choices[0].message.content is not None
        return res.choices[0].message.content



import PIL.ImageFilter
import PIL.ImageOps
import PIL.Image
import PIL.ImageEnhance
import numpy as np
from pydantic import Field
from nodetool.metadata.types import FolderRef
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import ImageRef
from nodetool.workflows.base_node import BaseNode
from nodetool.metadata.types import VideoRef
import tempfile


class SaveVideo(BaseNode):
    """
    Save a video to a file.
    video, save, file, output
    """

    value: VideoRef = VideoRef()
    folder: FolderRef = Field(
        default=FolderRef(), description="Name of the output folder."
    )
    name: str = Field(default="video", description="Name of the output video.")

    async def process(self, context: ProcessingContext) -> VideoRef:
        video = await context.asset_to_io(self.value)
        return await context.video_from_io(
            buffer=video,
            name=self.name,
            parent_id=self.folder.asset_id,
        )


class ExtractVideoFrames(BaseNode):
    """
    Extracts frames from a video file.
    video, frames, extract, sequence
    """

    video: VideoRef = Field(description="The input video to adjust the brightness for.")
    start: int = Field(default=0, description="The frame to start extracting from.")
    end: int = Field(default=-1, description="The frame to stop extracting from.")

    async def process(self, context: ProcessingContext) -> list[ImageRef]:
        import imageio.v3 as iio

        video_file = await context.asset_to_io(self.video)
        images = []
        with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp:
            temp.write(video_file.read())
            temp.flush()
            for i, frame in enumerate(iio.imiter(temp.name, plugin="pyav")):
                if i > self.start and (self.end == -1 or i < self.end):
                    img = PIL.Image.fromarray(frame)
                    img_ref = await context.image_from_pil(img)
                    images.append(img_ref)
        return images


class VideoFps(BaseNode):
    """
    Returns the frames per second (FPS) of a video file.
    video, analysis, frames, fps
    Outputs the numerical FPS value of the input video.
    """

    video: VideoRef = Field(description="The input video to adjust the brightness for.")

    async def process(self, context: ProcessingContext) -> float:
        import imageio.v3 as iio

        video_file = await context.asset_to_io(self.video)
        with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp:
            temp.write(video_file.read())
            temp.flush()
            return iio.immeta(temp.name, plugin="pyav")["fps"]


class FramesToVideo(BaseNode):
    """
    Combines a sequence of frames into a single video file.
    video, frames, combine, sequence
    Returns a video file from the provided frame sequence.
    """

    frames: list[ImageRef] = []
    fps: float = Field(default=30, description="The FPS of the output video.")

    async def process(self, context: ProcessingContext) -> VideoRef:
        import imageio.v3 as iio

        with tempfile.NamedTemporaryFile(suffix=".mp4", delete=True) as temp:
            out = iio.imopen(temp.name, "w", plugin="pyav")
            out.init_video_stream("vp9", fps=self.fps)
            for img_ref in self.frames:
                img = await context.image_to_pil(img_ref)
                out.write_frame(np.array(img))
            out.close()
            return await context.video_from_io(open(temp.name, "rb"))



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum

from nodetool.metadata.types import ImageRef


class Model(str, Enum):
    IMAGE_DENOISING = "Image Denoising"
    IMAGE_DEBLURRING__GOPRO = "Image Deblurring (GoPro)"
    IMAGE_DEBLURRING__REDS = "Image Deblurring (REDS)"
    IMAGE_DEBLURRING__REALBLUR_R = "Image Deblurring (RealBlur_R)"
    IMAGE_DEBLURRING__REALBLUR_J = "Image Deblurring (RealBlur_J)"
    IMAGE_DERAINING__RAIN_STREAK = "Image Deraining (Rain streak)"
    IMAGE_DERAINING__RAIN_DROP = "Image Deraining (Rain drop)"
    IMAGE_DEHAZING__INDOOR = "Image Dehazing (Indoor)"
    IMAGE_DEHAZING__OUTDOOR = "Image Dehazing (Outdoor)"
    IMAGE_ENHANCEMENT__LOW_LIGHT = "Image Enhancement (Low-light)"
    IMAGE_ENHANCEMENT__RETOUCHING = "Image Enhancement (Retouching)"


class Maxim(ReplicateNode):
    """Multi-Axis MLP for Image Processing"""

    @classmethod
    def replicate_model_id(cls):
        return "google-research/maxim:494ca4d578293b4b93945115601b6a38190519da18467556ca223d219c3af9f9"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/mgxm/716ffa94-41f7-46b4-a2f0-287f0e907f9c/output.png",
            "created_at": "2022-04-20T16:32:30.049391Z",
            "description": "Multi-Axis MLP for Image Processing",
            "github_url": "https://github.com/google-research/maxim",
            "license_url": "https://github.com/google-research/maxim/blob/main/LICENSE",
            "name": "maxim",
            "owner": "google-research",
            "paper_url": "https://arxiv.org/abs/2201.02973",
            "run_count": 406317,
            "url": "https://replicate.com/google-research/maxim",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="Input image.")
    model: Model = Field(description="Choose a model.", default=None)


class OldPhotosRestoration(ReplicateNode):
    """Bringing Old Photos Back to Life"""

    @classmethod
    def replicate_model_id(cls):
        return "microsoft/bringing-old-photos-back-to-life:c75db81db6cbd809d93cc3b7e7a088a351a3349c9fa02b6d393e35e0d51ba799"

    @classmethod
    def get_hardware(cls):
        return "Nvidia T4 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://replicate.delivery/mgxm/18b02ad9-ae81-4044-800e-65732df6bdc7/out.png",
            "created_at": "2021-09-11T14:44:30.681818Z",
            "description": "Bringing Old Photos Back to Life",
            "github_url": "https://github.com/microsoft/Bringing-Old-Photos-Back-to-Life",
            "license_url": "https://github.com/microsoft/Bringing-Old-Photos-Back-to-Life/blob/master/LICENSE",
            "name": "bringing-old-photos-back-to-life",
            "owner": "microsoft",
            "paper_url": "https://arxiv.org/abs/2004.09484",
            "run_count": 852778,
            "url": "https://replicate.com/microsoft/bringing-old-photos-back-to-life",
            "visibility": "public",
            "hardware": "Nvidia T4 GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    HR: bool = Field(
        title="Hr",
        description="whether the input image is high-resolution",
        default=False,
    )
    image: ImageRef = Field(default=ImageRef(), description="input image.")
    with_scratch: bool = Field(
        title="With Scratch",
        description="whether the input image is scratched",
        default=False,
    )


class CodeFormer(ReplicateNode):
    """Robust face restoration algorithm for old photos/AI-generated faces - (A40 GPU)"""

    @classmethod
    def replicate_model_id(cls):
        return "lucataco/codeformer:78f2bab438ab0ffc85a68cdfd316a2ecd3994b5dd26aa6b3d203357b45e5eb1b"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 (Large) GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/cf736d61-411f-4301-89b3-16aff1a02ed1/codeformer_logo.png",
            "created_at": "2023-09-06T04:10:50.158696Z",
            "description": "Robust face restoration algorithm for old photos/AI-generated faces - (A40 GPU)",
            "github_url": "https://github.com/sczhou/CodeFormer",
            "license_url": "https://github.com/sczhou/CodeFormer/blob/master/LICENSE",
            "name": "codeformer",
            "owner": "lucataco",
            "paper_url": "https://arxiv.org/abs/2206.11253",
            "run_count": 275732,
            "url": "https://replicate.com/lucataco/codeformer",
            "visibility": "public",
            "hardware": "Nvidia A40 (Large) GPU",
        }

    @classmethod
    def return_type(cls):
        return ImageRef

    image: ImageRef = Field(default=ImageRef(), description="Input image")
    upscale: int = Field(
        title="Upscale",
        description="The final upsampling scale of the image",
        default=2,
    )
    face_upsample: bool = Field(
        title="Face Upsample",
        description="Upsample restored faces for high-resolution AI-created images",
        default=True,
    )
    background_enhance: bool = Field(
        title="Background Enhance",
        description="Enhance background image with Real-ESRGAN",
        default=True,
    )
    codeformer_fidelity: float = Field(
        title="Codeformer Fidelity",
        description="Balance the quality (lower number) and fidelity (higher number).",
        ge=0.0,
        le=1.0,
        default=0.5,
    )



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class AnimeLineArt_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types()

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, resolution=512, **kwargs):
        from controlnet_aux.lineart_anime import LineartAnimeDetector

        model = LineartAnimeDetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution)
        del model
        return (out, )

NODE_CLASS_MAPPINGS = {
    "AnimeLineArtPreprocessor": AnimeLineArt_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "AnimeLineArtPreprocessor": "Anime Lineart"
}


from datetime import datetime
import uuid

from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid
from typing import Any


class Prediction(DBModel):
    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_predictions",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {"id": "S", "user_id": "S", "workflow_id": "S"},
            "global_secondary_indexes": {
                "nodetool_prediction_user_index": {
                    "user_id": "HASH",
                    "workflow_id": "RANGE",
                },
            },
        }

    id: str = DBField(hash_key=True)
    user_id: str = DBField()
    node_id: str = DBField(default="")
    provider: str = DBField(default="")
    model: str = DBField()
    workflow_id: str | None = DBField(default=None)
    error: str | None = DBField(default=None)
    logs: str | None = DBField(default=None)
    status: str = DBField(default="starting")
    created_at: datetime | None = DBField(default_factory=datetime.now)
    started_at: datetime | None = DBField(default=None)
    completed_at: datetime | None = DBField(default=None)
    cost: float | None = DBField(default=None)
    duration: float | None = DBField(default=None)
    hardware: str | None = DBField(default=None)

    @classmethod
    def create(
        cls,
        user_id: str,
        node_id: str,
        provider: str,
        model: str,
        workflow_id: str | None = None,
        status: str = "starting",
        cost: float | None = None,
        hardware: str | None = None,
        started_at: datetime | None = None,
    ):
        prediction = cls(
            id=create_time_ordered_uuid(),
            user_id=user_id,
            node_id=node_id,
            provider=provider,
            model=model,
            workflow_id=workflow_id,
            status=status,
            cost=cost,
            hardware=hardware,
            started_at=started_at,
        )
        prediction.save()
        return prediction

    @classmethod
    def find(cls, user_id: str, id: str):
        prediction = cls.get(id)
        if prediction is None or prediction.user_id != user_id:
            return None
        return prediction

    @classmethod
    def paginate(
        cls,
        user_id: str,
        workflow_id: str | None = None,
        limit: int = 100,
        start_key: str | None = None,
    ):
        if workflow_id is None:
            return cls.query(
                condition="user_id = :user_id",
                values={":user_id": user_id},
                index="nodetool_prediction_user_index",
                limit=limit,
                start_key=start_key,
            )
        else:
            return cls.query(
                condition="user_id = :user_id AND workflow_id = :workflow_id",
                values={":user_id": user_id, ":workflow_id": workflow_id},
                index="nodetool_prediction_user_index",
                limit=limit,
                start_key=start_key,
            )



# Openpose
# Original from CMU https://github.com/CMU-Perceptual-Computing-Lab/openpose
# 2nd Edited by https://github.com/Hzzone/pytorch-openpose
# 3rd Edited by ControlNet
# 4th Edited by ControlNet (added face and correct hands)
# 5th Edited by ControlNet (Improved JSON serialization/deserialization, and lots of bug fixs)
# This preprocessor is licensed by CMU for non-commercial use only.
import torch.utils.benchmark as benchmark
benchmark.timer()

import os
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import json
import torch
import numpy as np
from . import util
from .body import Body, BodyResult, Keypoint
from .hand import Hand
from .face import Face
from .types import PoseResult, HandResult, FaceResult
from huggingface_hub import hf_hub_download
from .wholebody import Wholebody
import warnings
from controlnet_aux.util import HWC3, resize_image_with_pad, common_input_validate, custom_hf_download
import cv2
from PIL import Image
from .animalpose import AnimalPoseImage

from typing import Tuple, List, Callable, Union, Optional

def draw_poses(poses: List[PoseResult], H, W, draw_body=True, draw_hand=True, draw_face=True):
    """
    Draw the detected poses on an empty canvas.

    Args:
        poses (List[PoseResult]): A list of PoseResult objects containing the detected poses.
        H (int): The height of the canvas.
        W (int): The width of the canvas.
        draw_body (bool, optional): Whether to draw body keypoints. Defaults to True.
        draw_hand (bool, optional): Whether to draw hand keypoints. Defaults to True.
        draw_face (bool, optional): Whether to draw face keypoints. Defaults to True.

    Returns:
        numpy.ndarray: A 3D numpy array representing the canvas with the drawn poses.
    """
    canvas = np.zeros(shape=(H, W, 3), dtype=np.uint8)

    for pose in poses:
        if draw_body:
            canvas = util.draw_bodypose(canvas, pose.body.keypoints)

        if draw_hand:
            canvas = util.draw_handpose(canvas, pose.left_hand)
            canvas = util.draw_handpose(canvas, pose.right_hand)

        if draw_face:
            canvas = util.draw_facepose(canvas, pose.face)

    return canvas


def decode_json_as_poses(json_string: str, normalize_coords: bool = False) -> Tuple[List[PoseResult], int, int]:
    """ Decode the json_string complying with the openpose JSON output format
    to poses that controlnet recognizes.
    https://github.com/CMU-Perceptual-Computing-Lab/openpose/blob/master/doc/02_output.md

    Args:
        json_string: The json string to decode.
        normalize_coords: Whether to normalize coordinates of each keypoint by canvas height/width.
                          `draw_pose` only accepts normalized keypoints. Set this param to True if
                          the input coords are not normalized.
    
    Returns:
        poses
        canvas_height
        canvas_width                      
    """
    pose_json = json.loads(json_string)
    height = pose_json['canvas_height']
    width = pose_json['canvas_width']

    def chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]
    
    def decompress_keypoints(numbers: Optional[List[float]]) -> Optional[List[Optional[Keypoint]]]:
        if not numbers:
            return None
        
        assert len(numbers) % 3 == 0

        def create_keypoint(x, y, c):
            if c < 1.0:
                return None
            keypoint = Keypoint(x, y)
            return keypoint

        return [
            create_keypoint(x, y, c)
            for x, y, c in chunks(numbers, n=3)
        ]
    
    return (
        [
            PoseResult(
                body=BodyResult(keypoints=decompress_keypoints(pose.get('pose_keypoints_2d'))),
                left_hand=decompress_keypoints(pose.get('hand_left_keypoints_2d')),
                right_hand=decompress_keypoints(pose.get('hand_right_keypoints_2d')),
                face=decompress_keypoints(pose.get('face_keypoints_2d'))
            )
            for pose in pose_json['people']
        ],
        height,
        width,
    )


def encode_poses_as_dict(poses: List[PoseResult], canvas_height: int, canvas_width: int) -> str:
    """ Encode the pose as a dict following openpose JSON output format:
    https://github.com/CMU-Perceptual-Computing-Lab/openpose/blob/master/doc/02_output.md
    """
    def compress_keypoints(keypoints: Union[List[Keypoint], None]) -> Union[List[float], None]:
        if not keypoints:
            return None
        
        return [
            value
            for keypoint in keypoints
            for value in (
                [float(keypoint.x), float(keypoint.y), 1.0]
                if keypoint is not None
                else [0.0, 0.0, 0.0]
            )
        ]

    return {
        'people': [
            {
                'pose_keypoints_2d': compress_keypoints(pose.body.keypoints),
                "face_keypoints_2d": compress_keypoints(pose.face),
                "hand_left_keypoints_2d": compress_keypoints(pose.left_hand),
                "hand_right_keypoints_2d":compress_keypoints(pose.right_hand),
            }
            for pose in poses
        ],
        'canvas_height': canvas_height,
        'canvas_width': canvas_width,
    }

global_cached_dwpose = Wholebody()

class DwposeDetector:
    """
    A class for detecting human poses in images using the Dwpose model.

    Attributes:
        model_dir (str): Path to the directory where the pose models are stored.
    """
    def __init__(self, dw_pose_estimation):
        self.dw_pose_estimation = dw_pose_estimation
    
    @classmethod
    def from_pretrained(cls, pretrained_model_or_path, pretrained_det_model_or_path=None, det_filename=None, pose_filename=None, torchscript_device="cuda"):
        global global_cached_dwpose
        pretrained_det_model_or_path = pretrained_det_model_or_path or pretrained_model_or_path
        det_filename = det_filename or "yolox_l.onnx"
        pose_filename = pose_filename or "dw-ll_ucoco_384.onnx"
        det_model_path = custom_hf_download(pretrained_det_model_or_path, det_filename)
        pose_model_path = custom_hf_download(pretrained_model_or_path, pose_filename)
        
        print(f"\nDWPose: Using {det_filename} for bbox detection and {pose_filename} for pose estimation")
        if global_cached_dwpose.det is None or global_cached_dwpose.det_filename != det_filename:
            t = Wholebody(det_model_path, None, torchscript_device=torchscript_device)
            t.pose = global_cached_dwpose.pose
            t.pose_filename = global_cached_dwpose.pose
            global_cached_dwpose = t
        
        if global_cached_dwpose.pose is None or global_cached_dwpose.pose_filename != pose_filename:
            t = Wholebody(None, pose_model_path, torchscript_device=torchscript_device)
            t.det = global_cached_dwpose.det
            t.det_filename = global_cached_dwpose.det_filename
            global_cached_dwpose = t
        return cls(global_cached_dwpose)

    def detect_poses(self, oriImg) -> List[PoseResult]:
        with torch.no_grad():
            keypoints_info = self.dw_pose_estimation(oriImg.copy())
            return Wholebody.format_result(keypoints_info)
    
    def __call__(self, input_image, detect_resolution=512, include_body=True, include_hand=False, include_face=False, hand_and_face=None, output_type="pil", image_and_json=False, upscale_method="INTER_CUBIC", **kwargs):
        if hand_and_face is not None:
            warnings.warn("hand_and_face is deprecated. Use include_hand and include_face instead.", DeprecationWarning)
            include_hand = hand_and_face
            include_face = hand_and_face

        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        
        poses = self.detect_poses(input_image)
        canvas = draw_poses(poses, input_image.shape[0], input_image.shape[1], draw_body=include_body, draw_hand=include_hand, draw_face=include_face) 
        detected_map = HWC3(remove_pad(canvas))

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
        
        if image_and_json:
            return (detected_map, encode_poses_as_dict(poses, detected_map.shape[0], detected_map.shape[1]))
        
        return detected_map

global_cached_animalpose = AnimalPoseImage()
class AnimalposeDetector:
    """
    A class for detecting animal poses in images using the RTMPose AP10k model.

    Attributes:
        model_dir (str): Path to the directory where the pose models are stored.
    """
    def __init__(self, animal_pose_estimation):
        self.animal_pose_estimation = animal_pose_estimation
    
    @classmethod
    def from_pretrained(cls, pretrained_model_or_path, pretrained_det_model_or_path=None, det_filename="yolox_l.onnx", pose_filename="dw-ll_ucoco_384.onnx", torchscript_device="cuda"):
        global global_cached_animalpose
        det_model_path = custom_hf_download(pretrained_det_model_or_path, det_filename)
        pose_model_path = custom_hf_download(pretrained_model_or_path, pose_filename)
        
        print(f"\nAnimalPose: Using {det_filename} for bbox detection and {pose_filename} for pose estimation")
        if global_cached_animalpose.det is None or global_cached_animalpose.det_filename != det_filename:
            t = AnimalPoseImage(det_model_path, None, torchscript_device=torchscript_device)
            t.pose = global_cached_animalpose.pose
            t.pose_filename = global_cached_animalpose.pose
            global_cached_animalpose = t
        
        if global_cached_animalpose.pose is None or global_cached_animalpose.pose_filename != pose_filename:
            t = AnimalPoseImage(None, pose_model_path, torchscript_device=torchscript_device)
            t.det = global_cached_animalpose.det
            t.det_filename = global_cached_animalpose.det_filename
            global_cached_animalpose = t
        return cls(global_cached_animalpose)
    
    def __call__(self, input_image, detect_resolution=512, output_type="pil", image_and_json=False, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        detected_map, openpose_dict = self.animal_pose_estimation(input_image)
        detected_map = remove_pad(detected_map)

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
        
        if image_and_json:
            return (detected_map, openpose_dict)

        return detected_map


import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import resize_image_with_pad, common_input_validate, HWC3, custom_hf_download, MESH_GRAPHORMER_MODEL_NAME
from controlnet_aux.mesh_graphormer.pipeline import MeshGraphormerMediapipe, args

class MeshGraphormerDetector:
    def __init__(self, pipeline):
        self.pipeline = pipeline

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=MESH_GRAPHORMER_MODEL_NAME, filename="graphormer_hand_state_dict.bin", hrnet_filename="hrnetv2_w64_imagenet_pretrained.pth"):
        args.resume_checkpoint = custom_hf_download(pretrained_model_or_path, filename)
        args.hrnet_checkpoint = custom_hf_download(pretrained_model_or_path, hrnet_filename)
        pipeline = MeshGraphormerMediapipe(args)
        return cls(pipeline)
    
    def to(self, device):
        self.pipeline._model.to(device)
        self.pipeline.mano_model.to(device)
        self.pipeline.mano_model.layer.to(device)
        return self

    def __call__(self, input_image=None, mask_bbox_padding=30, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)

        depth_map, mask, info = self.pipeline.get_depth(input_image, mask_bbox_padding)
        if depth_map is None:
            depth_map = np.zeros_like(input_image)
            mask = np.zeros_like(input_image)

        #The hand is small
        depth_map, mask = HWC3(depth_map), HWC3(mask)
        depth_map, remove_pad = resize_image_with_pad(depth_map, detect_resolution, upscale_method)
        depth_map = remove_pad(depth_map)
        if output_type == "pil":
            depth_map = Image.fromarray(depth_map)
            mask = Image.fromarray(mask)
            
        return depth_map, mask, info



# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/facebookresearch/detr/blob/master/util/misc.py
"""
Misc functions, including distributed helpers.

Mostly copy-paste from torchvision references.
"""
from typing import List, Optional

import torch
import torch.distributed as dist
import torchvision
from torch import Tensor
import warnings
import torch.nn.functional as F
import math

def inverse_sigmoid(x, eps=1e-3):
    x = x.clamp(min=0, max=1)
    x1 = x.clamp(min=eps)
    x2 = (1 - x).clamp(min=eps)
    return torch.log(x1/x2)

def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor

def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    # type: (Tensor, float, float, float, float) -> Tensor
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)

def resize(input,
           size=None,
           scale_factor=None,
           mode='nearest',
           align_corners=None,
           warning=True):
    if warning:
        if size is not None and align_corners:
            input_h, input_w = tuple(int(x) for x in input.shape[2:])
            output_h, output_w = tuple(int(x) for x in size)
            if output_h > input_h or output_w > output_h:
                if ((output_h > 1 and output_w > 1 and input_h > 1
                     and input_w > 1) and (output_h - 1) % (input_h - 1)
                        and (output_w - 1) % (input_w - 1)):
                    warnings.warn(
                        f'When align_corners={align_corners}, '
                        'the output would more aligned if '
                        f'input size {(input_h, input_w)} is `x+1` and '
                        f'out size {(output_h, output_w)} is `nx+1`')
    if isinstance(size, torch.Size):
        size = tuple(int(x) for x in size)
    return F.interpolate(input, size, scale_factor, mode, align_corners)

def _max_by_axis(the_list):
    # type: (List[List[int]]) -> List[int]
    maxes = the_list[0]
    for sublist in the_list[1:]:
        for index, item in enumerate(sublist):
            maxes[index] = max(maxes[index], item)
    return maxes


class NestedTensor(object):
    def __init__(self, tensors, mask: Optional[Tensor]):
        self.tensors = tensors
        self.mask = mask

    def to(self, device):
        # type: (Device) -> NestedTensor # noqa
        cast_tensor = self.tensors.to(device)
        mask = self.mask
        if mask is not None:
            assert mask is not None
            cast_mask = mask.to(device)
        else:
            cast_mask = None
        return NestedTensor(cast_tensor, cast_mask)

    def decompose(self):
        return self.tensors, self.mask

    def __repr__(self):
        return str(self.tensors)


def nested_tensor_from_tensor_list(tensor_list: List[Tensor]):
    # TODO make this more general
    if tensor_list[0].ndim == 3:
        if torchvision._is_tracing():
            # nested_tensor_from_tensor_list() does not export well to ONNX
            # call _onnx_nested_tensor_from_tensor_list() instead
            return _onnx_nested_tensor_from_tensor_list(tensor_list)

        # TODO make it support different-sized images
        max_size = _max_by_axis([list(img.shape) for img in tensor_list])
        # min_size = tuple(min(s) for s in zip(*[img.shape for img in tensor_list]))
        batch_shape = [len(tensor_list)] + max_size
        b, c, h, w = batch_shape
        dtype = tensor_list[0].dtype
        device = tensor_list[0].device
        tensor = torch.zeros(batch_shape, dtype=dtype, device=device)
        mask = torch.ones((b, h, w), dtype=torch.bool, device=device)
        for img, pad_img, m in zip(tensor_list, tensor, mask):
            pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
            m[: img.shape[1], : img.shape[2]] = False
    else:
        raise ValueError("not supported")
    return NestedTensor(tensor, mask)


# _onnx_nested_tensor_from_tensor_list() is an implementation of
# nested_tensor_from_tensor_list() that is supported by ONNX tracing.
@torch.jit.unused
def _onnx_nested_tensor_from_tensor_list(tensor_list: List[Tensor]) -> NestedTensor:
    max_size = []
    for i in range(tensor_list[0].dim()):
        max_size_i = torch.max(
            torch.stack([img.shape[i] for img in tensor_list]).to(torch.float32)
        ).to(torch.int64)
        max_size.append(max_size_i)
    max_size = tuple(max_size)

    # work around for
    # pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
    # m[: img.shape[1], :img.shape[2]] = False
    # which is not yet supported in onnx
    padded_imgs = []
    padded_masks = []
    for img in tensor_list:
        padding = [(s1 - s2) for s1, s2 in zip(max_size, tuple(img.shape))]
        padded_img = torch.nn.functional.pad(img, (0, padding[2], 0, padding[1], 0, padding[0]))
        padded_imgs.append(padded_img)

        m = torch.zeros_like(img[0], dtype=torch.int, device=img.device)
        padded_mask = torch.nn.functional.pad(m, (0, padding[2], 0, padding[1]), "constant", 1)
        padded_masks.append(padded_mask.to(torch.bool))

    tensor = torch.stack(padded_imgs)
    mask = torch.stack(padded_masks)

    return NestedTensor(tensor, mask=mask)


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True



# ------------------------------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------------------------------
# Modified from https://github.com/chengdazhi/Deformable-Convolution-V2-PyTorch/tree/pytorch_1.0.0
# ------------------------------------------------------------------------------------------------

# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/fundamentalvision/Deformable-DETR

import os
import glob

import torch

from torch.utils.cpp_extension import CUDA_HOME
from torch.utils.cpp_extension import CppExtension
from torch.utils.cpp_extension import CUDAExtension

from setuptools import find_packages
from setuptools import setup

requirements = ["torch", "torchvision"]

def get_extensions():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    extensions_dir = os.path.join(this_dir, "src")

    main_file = glob.glob(os.path.join(extensions_dir, "*.cpp"))
    source_cpu = glob.glob(os.path.join(extensions_dir, "cpu", "*.cpp"))
    source_cuda = glob.glob(os.path.join(extensions_dir, "cuda", "*.cu"))

    sources = main_file + source_cpu
    extension = CppExtension
    extra_compile_args = {"cxx": []}
    define_macros = []

    # Force cuda since torch ask for a device, not if cuda is in fact available.
    if (os.environ.get('FORCE_CUDA') or torch.cuda.is_available()) and CUDA_HOME is not None:
        extension = CUDAExtension
        sources += source_cuda
        define_macros += [("WITH_CUDA", None)]
        extra_compile_args["nvcc"] = [
            "-DCUDA_HAS_FP16=1",
            "-D__CUDA_NO_HALF_OPERATORS__",
            "-D__CUDA_NO_HALF_CONVERSIONS__",
            "-D__CUDA_NO_HALF2_OPERATORS__",
        ]
    else:
        if CUDA_HOME is None:
            raise NotImplementedError('CUDA_HOME is None. Please set environment variable CUDA_HOME.')
        else:
            raise NotImplementedError('No CUDA runtime is found. Please set FORCE_CUDA=1 or test it by running torch.cuda.is_available().')

    sources = [os.path.join(extensions_dir, s) for s in sources]
    include_dirs = [extensions_dir]
    ext_modules = [
        extension(
            "MultiScaleDeformableAttention",
            sources,
            include_dirs=include_dirs,
            define_macros=define_macros,
            extra_compile_args=extra_compile_args,
        )
    ]
    return ext_modules

setup(
    name="MultiScaleDeformableAttention",
    version="1.0",
    author="Weijie Su",
    url="https://github.com/fundamentalvision/Deformable-DETR",
    description="PyTorch Wrapper for CUDA Functions of Multi-Scale Deformable Attention",
    packages=find_packages(exclude=("configs", "tests",)),
    ext_modules=get_extensions(),
    cmdclass={"build_ext": torch.utils.cpp_extension.BuildExtension},
)



from .backbone.swin import D2SwinTransformer
from .backbone.dinat import D2DiNAT
from .pixel_decoder.fpn import BasePixelDecoder
from .pixel_decoder.msdeformattn import MSDeformAttnPixelDecoder
from .meta_arch.oneformer_head import OneFormerHead



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Llama3_70B(GraphNode):
    top_k: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Prompt')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The maximum number of tokens the model should generate as output.')
    min_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The minimum number of tokens the model should generate as output.')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='The value used to modulate the next token probabilities.')
    prompt_template: str | GraphNode | tuple[GraphNode, str] = Field(default='{prompt}', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.')
    presence_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=1.15, description='Presence penalty')
    frequency_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Frequency penalty')
    @classmethod
    def get_node_type(cls): return "replicate.text.generate.Llama3_70B"



class Llama3_70B_Instruct(GraphNode):
    top_k: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Prompt')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The maximum number of tokens the model should generate as output.')
    min_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The minimum number of tokens the model should generate as output.')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='The value used to modulate the next token probabilities.')
    prompt_template: str | GraphNode | tuple[GraphNode, str] = Field(default='{prompt}', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.')
    presence_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=1.15, description='Presence penalty')
    frequency_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Frequency penalty')
    @classmethod
    def get_node_type(cls): return "replicate.text.generate.Llama3_70B_Instruct"



class Llama3_8B(GraphNode):
    top_k: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Prompt')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The maximum number of tokens the model should generate as output.')
    min_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The minimum number of tokens the model should generate as output.')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='The value used to modulate the next token probabilities.')
    prompt_template: str | GraphNode | tuple[GraphNode, str] = Field(default='{prompt}', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.')
    presence_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=1.15, description='Presence penalty')
    frequency_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Frequency penalty')
    @classmethod
    def get_node_type(cls): return "replicate.text.generate.Llama3_8B"



class Llama3_8B_Instruct(GraphNode):
    top_k: int | GraphNode | tuple[GraphNode, str] = Field(default=50, description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).')
    top_p: float | GraphNode | tuple[GraphNode, str] = Field(default=0.9, description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Prompt')
    max_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='The maximum number of tokens the model should generate as output.')
    min_tokens: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The minimum number of tokens the model should generate as output.')
    temperature: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='The value used to modulate the next token probabilities.')
    prompt_template: str | GraphNode | tuple[GraphNode, str] = Field(default='{prompt}', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.')
    presence_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=1.15, description='Presence penalty')
    frequency_penalty: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Frequency penalty')
    @classmethod
    def get_node_type(cls): return "replicate.text.generate.Llama3_8B_Instruct"





from enum import Enum
import random
from typing import Any
import numpy as np
from pydantic import Field, validator
from nodetool.metadata.types import ImageRef
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import ImageTensor, Mask
from nodetool.common.comfy_node import ComfyNode


class ImageCompositeMasked(ComfyNode):
    destination: ImageTensor = Field(
        default=ImageTensor(), description="The destination image."
    )
    source: ImageTensor = Field(default=ImageTensor(), description="The source image.")
    x: int = Field(default=0, description="The x position.")
    y: int = Field(default=0, description="The y position.")
    resize_source: bool = Field(
        default=False, description="Whether to resize the source."
    )
    mask: Mask = Field(None, description="The mask to use.")

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


class LoadImage(ComfyNode):
    """
    The Load Image node can be used to to load an image. Images can be uploaded in the asset manager or by dropping an image onto the node. Once the image has been uploaded they can be selected inside the node.
    """

    image: ImageRef = Field(default=ImageRef(), description="The image to load.")
    upload: str = Field(default="", description="unused")

    @validator("image", pre=True)
    def validate_image(cls, v):
        if isinstance(v, str):
            v = ImageRef(uri=v)
        return v

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}

    def assign_property(self, name: str, value: Any):
        if name == "image" and isinstance(value, str):
            # comfy import with local filename
            return
        super().assign_property(name, value)


class ColorChannel(str, Enum):
    ALPHA = "alpha"
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class LoadImageMask(ComfyNode):
    image: ImageTensor = Field(default=ImageTensor(), description="The image to load.")
    channel: ColorChannel = Field(
        default=ColorChannel.ALPHA, description="The color channel to use."
    )

    @validator("image", pre=True)
    def validate_image(cls, v):
        if isinstance(v, str):
            v = ImageRef(uri=v)
        return v

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class SaveImage(ComfyNode):
    """
    The Save Image node can be used to save images. To simply preview an image inside the node graph use the Preview Image node. It can be hard to keep track of all the images that you generate. To help with organizing your images you can pass specially formatted strings to an output node with a file_prefix widget.
    """

    images: ImageTensor = Field(default=ImageTensor(), description="The image to save.")
    filename_prefix: str = Field(
        default="",
        description="The prefix for the filename where the image will be saved.",
    )

    async def process(self, context: ProcessingContext):
        image = self.images[0]  # type: ignore
        i = 255.0 * image.cpu().detach().numpy()  # type: ignore
        img = np.clip(i, 0, 255).astype(np.uint8)
        rand = "".join(random.choice("abcdefghijklmnopqrstupvxyz") for x in range(5))
        name = f"{self.filename_prefix}_{rand}.png"
        ref = await context.image_from_numpy(img, name=name)
        return {"image": ref}

    @classmethod
    def return_type(cls):
        return {"image": ImageRef}


class PreviewImage(SaveImage):
    """
    The Preview Image node can be used to preview images inside the node graph.
    """

    pass


class ImageInvert(ComfyNode):
    """
    The Invert Image node can be used to to invert the colors of an image.
    """

    image: ImageTensor = Field(
        default=ImageTensor(), description="The image to invert."
    )

    @classmethod
    def get_title(cls):
        return "Invert Image"

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


class ImageBatch(ComfyNode):
    image1: ImageTensor = Field(
        default=ImageTensor(), description="The image to batch."
    )
    image2: ImageTensor = Field(
        default=ImageTensor(), description="The image to batch."
    )

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


class ImagePadForOutpaint(ComfyNode):
    """
    The Pad Image for Outpainting node can be used to to add padding to an image for outpainting. This image can then be given to an inpaint diffusion model via the VAE Encode for Inpainting.
    """

    image: ImageTensor = Field(default=ImageTensor(), description="The image to pad.")
    left: int = Field(default=0, description="The padding size on the left side.")
    top: int = Field(default=0, description="The padding size on the top side.")
    right: int = Field(default=0, description="The padding size on the right side.")
    bottom: int = Field(default=0, description="The padding size on the bottom side.")
    feathering: int = Field(
        default=40,
        description="The feathering value for softening the edges of the padding.",
    )

    @classmethod
    def get_title(cls):
        return "Pad Image for Outpainting"

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor, "mask": Mask}



from pydantic import Field
import PIL.Image
import numpy as np
from nodetool.metadata.types import AssetRef, Tensor, TextRef
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.workflows.base_node import BaseNode
from nodetool.metadata.types import ImageRef
from nodetool.metadata.types import FolderRef
from nodetool.workflows.types import NodeProgress


class NearestNeighbors(BaseNode):
    """
    Stores input embeddings in a database and retrieves the nearest neighbors for a query embedding.
    """

    documents: list[Tensor] = Field(
        default=[], description="The list of documents to search"
    )
    query: Tensor = Field(
        default=Tensor(),
        description="The query to search for",
    )
    n_neighbors: int = Field(default=1, description="The number of neighbors to return")

    @classmethod
    def return_type(cls):
        return {
            "distances": list[float],
            "indices": list[int],
        }

    async def process(self, context: ProcessingContext):
        from sklearn.neighbors import NearestNeighbors

        embeddings = np.array([e.to_numpy() for e in self.documents])
        nbrs = NearestNeighbors(n_neighbors=self.n_neighbors).fit(embeddings)
        distances, indices = nbrs.kneighbors(self.query.to_numpy().reshape(1, -1))
        return {
            "distances": distances[0].tolist(),
            "indices": indices[0].tolist(),
        }


class ChromaNode(BaseNode):

    def get_or_create_collection(self, context: ProcessingContext, name: str):
        """
        Get or create a collection with the given name.

        Args:
            context: The processing context.
            name: The name of the collection to get or create.
        """
        from chromadb.utils.embedding_functions import OpenCLIPEmbeddingFunction
        from chromadb.utils.data_loaders import ImageLoader

        embedding_function = OpenCLIPEmbeddingFunction()
        data_loader = ImageLoader()
        client = context.get_chroma_client()

        return client.get_or_create_collection(
            name=name, embedding_function=embedding_function, data_loader=data_loader
        )

    async def load_results(
        self, context: ProcessingContext, ids: list[str]
    ) -> list[AssetRef]:
        asset_refs = []
        for id in ids:
            asset = await context.find_asset(str(id))
            if asset.content_type.startswith("image"):
                uri = await context.get_asset_url(asset.id)
                ref = ImageRef(asset_id=asset.id, uri=uri)
                asset_refs.append(ref)
            if asset.content_type.startswith("text"):
                uri = await context.get_asset_url(asset.id)
                ref = TextRef(asset_id=asset.id, uri=uri)
                asset_refs.append(ref)
        return asset_refs


MAX_INDEX_SIZE = 1000


class IndexFolder(ChromaNode):
    """
    Index all the assets in a folder.
    """

    folder: FolderRef = Field(default=FolderRef(), description="The folder to index")

    async def process(self, context: ProcessingContext) -> FolderRef:
        if not self.folder.asset_id:
            raise ValueError("The folder needs to be selected")
        collection = self.get_or_create_collection(context, self.folder.asset_id)

        asset_list = await context.paginate_assets(self.folder.asset_id, MAX_INDEX_SIZE)
        total = len(asset_list.assets)
        context.post_message(NodeProgress(node_id=self._id, progress=0, total=total))

        for i, asset in enumerate(asset_list.assets):
            res = collection.get(asset.id)

            if len(res["ids"]) == 0:
                if asset.content_type.startswith("text"):
                    text = await context.download_asset(asset.id)
                    collection.add(
                        ids=[asset.id], documents=[text.read().decode("utf-8")]
                    )
                if asset.content_type.startswith("image"):
                    image_io = await context.download_asset(asset.id)
                    image = PIL.Image.open(image_io)
                    collection.add(ids=[asset.id], images=[np.array(image)])
            context.post_message(
                NodeProgress(node_id=self._id, progress=i, total=total)
            )

        return self.folder


class QueryImage(ChromaNode):
    """
    Query the index for similar images.
    """

    folder: FolderRef = Field(default=FolderRef(), description="The folder to query")
    image: ImageRef = Field(default=ImageRef(), description="The image to query")
    n_results: int = Field(default=1, description="The number of results to return")

    async def process(self, context: ProcessingContext) -> list[AssetRef]:
        if not self.folder.asset_id:
            raise ValueError("The folder needs to be selected")
        if not self.image.asset_id:
            raise ValueError("The image needs to be selected")

        collection = self.get_or_create_collection(context, self.folder.asset_id)
        image_io = await context.download_asset(self.image.asset_id)
        image = PIL.Image.open(image_io)
        result = collection.query(
            query_images=[np.array(image)], n_results=self.n_results
        )
        ids = [str(id) for id in result["ids"][0]]

        return await self.load_results(context, ids)


class QueryText(ChromaNode):
    """
    Query the index for similar text.
    """

    folder: FolderRef = Field(default=FolderRef(), description="The folder to query")
    text: TextRef = Field(default=TextRef(), description="The text to query")
    n_results: int = Field(default=1, description="The number of results to return")

    async def process(self, context: ProcessingContext) -> list[AssetRef]:
        if not self.folder.asset_id:
            raise ValueError("The folder needs to be selected")
        if not self.text.asset_id:
            raise ValueError("The text needs to be selected")

        collection = self.get_or_create_collection(context, self.folder.asset_id)
        text_io = await context.download_asset(self.text.asset_id)
        text = text_io.read().decode("utf-8")
        result = collection.query(query_texts=[text], n_results=self.n_results)
        ids = [str(id) for id in result["ids"][0]]

        return await self.load_results(context, ids)



import asyncio
import json
from typing import Any

from pydantic import BaseModel
from nodetool.common.environment import Environment
from nodetool.metadata.types import FunctionModel, Message, Task, ToolCall
from nodetool.models.workflow import Workflow
from nodetool.workflows.base_node import get_node_class
from nodetool.workflows.graph import Graph
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.api.types.chat import MessageCreateRequest, MessageList, TaskCreateRequest
from openai.types.shared_params import FunctionDefinition
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai._types import NotGiven
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam

from nodetool.workflows.run_job_request import RunJobRequest
from nodetool.workflows.run_workflow import run_workflow


def sanitize_node_name(node_name: str) -> str:
    """
    Sanitize a node name.

    Args:
        node_name (str): The node name.

    Returns:
        str: The sanitized node name.
    """
    return node_name.replace(".", "_")


def desanitize_node_name(node_name: str) -> str:
    """
    Desanitize a node name.

    Args:
        node_name (str): The node name.

    Returns:
        str: The desanitized node name.
    """
    return node_name.replace("_", ".")


NODE_PREFIX = "node__"
WORKFLOW_PREFIX = "workflow__"


def function_tool_from_node(node_name: str):
    """
    Create a function tool from a node definition.

    A node is technically a function, where the parameters are the properties of the node.
    The interface of the function is the JSON schema of the node.

    Args:
        node_name (str): The name of the node.
    """
    node_type = get_node_class(node_name)
    if node_type is None:
        raise ValueError(f"Node {node_name} does not exist")

    function_definition = FunctionDefinition(
        name=NODE_PREFIX + sanitize_node_name(node_name),
        description=node_type.get_description(),
        parameters=node_type.get_json_schema(),
    )
    return ChatCompletionToolParam(function=function_definition, type="function")


async def function_tool_from_workflow(context: ProcessingContext, workflow_id: str):
    workflow = await context.get_workflow(workflow_id)

    if workflow is None:
        raise ValueError(f"Workflow {workflow_id} does not exist")

    graph = Graph.from_dict(workflow.graph.model_dump())

    parameters = graph.get_input_schema()

    function_definition = FunctionDefinition(
        name=WORKFLOW_PREFIX + workflow_id,
        description=workflow.description or "",
        parameters=parameters,
    )
    return ChatCompletionToolParam(function=function_definition, type="function")


def create_task_tool():
    function_definition = FunctionDefinition(
        name="create_task",
        description="Create a task for an agent to be executed.",
        parameters={
            "type": "object",
            "properties": {
                "task_type": {
                    "type": "string",
                    "description": "The type of task. Describes the output of the task. For example, 'generate_text' or 'generate_image'.",
                    "enum": ["generate_text", "generate_image"],
                },
                "name": {
                    "type": "string",
                    "description": "The name of the task.",
                },
                "instructions": {
                    "type": "string",
                    "description": "Specific instructions for the agent to execute. For example, 'Generate a summary of the article.'",
                },
                "dependencies": {
                    "type": "array",
                    "description": "The dependencies of the task. Output of these tasks will be passed as context to the task.",
                    "items": {
                        "type": "string",
                        "description": "The name of the dependent task.",
                    },
                },
            },
        },
    )
    return ChatCompletionToolParam(function=function_definition, type="function")


def default_serializer(obj: Any) -> dict:
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    raise TypeError("Type not serializable")


async def process_node_function(
    context: ProcessingContext, node_type: str, params: dict
):
    """
    Process a node function with the given parameters.

    Args:
        context (ProcessingContext): The processing context.
        node_type (str): The node type.
        params (dict): The parameters passed to the node.
    """
    node_class = get_node_class(node_type)
    if node_class is None:
        raise ValueError(f"Node {node_type} does not exist")

    print(f"Processing node {node_type} with params {params}")

    node = node_class(id="")

    for key, value in params.items():
        node.assign_property(key, value)

    res = await node.process(context)
    out = await node.convert_output(context, res)
    return out


async def process_workflow_function(
    context: ProcessingContext, workflow_id: str, params: dict
) -> Any:
    """
    Process a workflow with the given parameters.
    If the node returns a prediction, wait for the prediction to complete.

    Args:
        context (ProcessingContext): The processing context.
        name (str): The workflow_id
        params (dict): The parameters passed to the workflow.
    """
    workflow = Workflow.get(workflow_id)
    if workflow is None:
        raise ValueError(f"Workflow {workflow_id} does not exist")

    req = RunJobRequest(
        user_id=context.user_id,
        auth_token=context.auth_token,
        graph=workflow.get_api_graph(),
        params=params,
    )

    output = {}

    for msg_json in run_workflow(req):
        msg = json.loads(msg_json)
        if msg["type"] == "node_progress":
            print(f"{msg['node_id']} -> {msg['progress']}\n")
        if msg["type"] == "node_update":
            print(f"{msg['node_name']} -> {msg['status']}\n")
        if msg["type"] == "error":
            raise Exception(msg["error"])
        if msg["type"] == "workflow_update":
            output = msg["result"]

    return {
        k: v.model_dump() if isinstance(v, BaseModel) else v for k, v in output.items()
    }


async def create_task(context: ProcessingContext, thread_id: str, params: dict):
    """
    Create a task for an agent to be executed.

    Args:
        context (ProcessingContext): The processing context.
        thread_id (str): The ID of the thread.
        user_id (str): The ID of the user.
        params (dict): The parameters of the task.
    """
    task = await context.create_task(
        TaskCreateRequest(
            thread_id=thread_id,
            task_type=params["task_type"],
            name=params["name"],
            instructions=params["instructions"],
            dependencies=params.get("dependencies", []),
        )
    )
    return task, {
        "type": "task",
        "task_type": task.task_type,
        "id": task.id,
        "name": task.name,
        "instructions": task.instructions,
        "dependencies": task.dependencies,
    }


async def create_completion(
    context: ProcessingContext,
    model: FunctionModel,
    messages: list[ChatCompletionMessageParam],
    tools: list[ChatCompletionToolParam],
    **kwargs,
) -> ChatCompletion:
    print(f"Creating completion with model {model.name}")

    if model.name.startswith("gpt"):
        client = Environment.get_openai_client()
        completion = await client.chat.completions.create(
            model=model.name, messages=messages, tools=tools if len(tools) > 0 else NotGiven()  # type: ignore
        )
    else:
        llm = context.load_llama_model(
            name=model.name,
            n_gpu_layers=kwargs.pop("n_gpu_layers", 0),
        )

        completion = ChatCompletion(
            **llm.create_chat_completion(
                messages=messages,  # type: ignore
                tools=tools,  # type: ignore
                **kwargs,
            )
        )

    return completion


async def process_messages(
    context: ProcessingContext,
    messages: list[Message],
    model: FunctionModel,
    thread_id: str,
    workflow_ids: list[str] = [],
    node_types: list[str] = [],
    can_create_tasks: bool = True,
    **kwargs,
) -> tuple[list[Message], list[Task], list[ToolCall]]:
    """
    Process a message in a thread.

    Args:
        context (ProcessingContext): The processing context.
        messages (list[Message]): The messages in the thread.
        model (str): The model to use for the completion.
        thread_id (str): The ID of the thread.
        workflow_ids (list[str]): The workflow IDs to use as tools.
        node_types (list[str]): The node types to use as tools.
    """

    messages_for_request = [
        {"role": str(message.role), "content": message.content} for message in messages
    ]

    tasks = []
    tools = []
    messages_to_return = []

    if can_create_tasks:
        tools.append(create_task_tool())
    tools += await asyncio.gather(
        *[
            function_tool_from_workflow(context, workflow_id)
            for workflow_id in workflow_ids
        ]
    )
    tools += [function_tool_from_node(node_type) for node_type in node_types]

    completion = await create_completion(
        context,
        model,
        messages_for_request,  # type: ignore
        tools,
        **kwargs,
    )

    response_message = completion.choices[0].message  # type: ignore
    messages_for_request.append(response_message)  # type: ignore
    assistant_message = await context.create_message(
        MessageCreateRequest(
            thread_id=thread_id,
            user_id=context.user_id,
            role="assistant",
            content=response_message.content,
            tool_calls=(
                [
                    {
                        "function": {
                            "name": c.function.name,
                            "arguments": c.function.arguments,
                        }
                    }
                    for c in response_message.tool_calls
                ]
                if response_message.tool_calls
                else None
            ),
        )
    )

    messages_to_return.append(assistant_message)

    async def run_tool(tool_call: ChatCompletionMessageToolCall):
        print(f"******* [[TOOL]] {tool_call.function.name} *******")

        function_name = tool_call.function.name.strip()
        function_args = json.loads(tool_call.function.arguments)

        if function_name == "create_task":
            task, function_response = await create_task(
                context, thread_id, function_args
            )
            tasks.append(task)
        elif function_name.startswith(WORKFLOW_PREFIX):
            function_response = await process_workflow_function(
                context, function_name[len(WORKFLOW_PREFIX) :], function_args
            )
        elif function_name.startswith(NODE_PREFIX):
            function_response = await process_node_function(
                context,
                desanitize_node_name(function_name[len(NODE_PREFIX) :]),
                function_args,
            )
        else:
            raise ValueError(f"Unknown function type {function_name}")

        content = json.dumps(function_response, default=default_serializer)
        message = await context.create_message(
            MessageCreateRequest(
                thread_id=thread_id,
                user_id=context.user_id,
                role="tool",
                tool_call_id=tool_call.id,
                name=function_name,
                content=content,
            )
        )

        call = ToolCall(
            function_name=function_name,
            function_args=function_args,
            function_response=function_response,
        )
        return message, call

    if response_message.tool_calls:
        results = await asyncio.gather(
            *[run_tool(tool_call) for tool_call in response_message.tool_calls]
        )
        tool_messages = [result[0] for result in results]
        tool_calls = [result[1] for result in results]
        for tool_message in tool_messages:
            messages_for_request.append(
                {
                    "tool_call_id": tool_message.tool_call_id or "",
                    "role": "tool",
                    "name": tool_message.name,
                    "content": tool_message.content or "",
                }
            )
        second_response = await create_completion(
            context=context,
            model=model,
            messages=messages_for_request,  # type: ignore
            tools=[],
            **kwargs,
        )
        print("SECOND RESPONSE", second_response)
        second_message = second_response.choices[0].message
        second_assistant_message = await context.create_message(
            MessageCreateRequest(
                thread_id=thread_id,
                user_id=context.user_id,
                role="assistant",
                content=second_message.content,
            )
        )
        messages_to_return.append(second_assistant_message)
    else:
        tool_calls = []

    return messages_to_return, tasks, tool_calls



from pydantic import BaseModel

from nodetool.common.content_types import CONTENT_TYPE_TO_EXTENSION
from nodetool.common.environment import Environment
from nodetool.models.asset import Asset as AssetModel


class Asset(BaseModel):
    id: str
    user_id: str
    workflow_id: str | None
    parent_id: str
    name: str
    status: str
    content_type: str
    created_at: str
    get_url: str | None
    thumb_url: str | None
    duration: float | None = None

    @property
    def file_extension(self) -> str:
        """
        Get the file extension of the asset.

        For example, if the content type is "image/jpeg", this will return "jpeg".
        """
        return (
            CONTENT_TYPE_TO_EXTENSION[self.content_type]  # type: ignore
            if self.content_type in CONTENT_TYPE_TO_EXTENSION
            else "bin"
        )

    @property
    def file_name(self) -> str:
        """
        Get the file name of the asset.
        """
        return f"{self.id}.{self.file_extension}"

    @classmethod
    def from_model(cls, asset: AssetModel):
        storage = Environment.get_asset_storage()
        if asset.content_type != "folder":
            get_url = storage.get_url(asset.file_name)
        else:
            get_url = None

        if asset.has_thumbnail:
            thumb_url = storage.get_url(asset.thumb_file_name)
        else:
            thumb_url = None

        return cls(
            id=asset.id,
            user_id=asset.user_id,
            workflow_id=asset.workflow_id,
            parent_id=asset.parent_id,
            name=asset.name,
            status=asset.status,
            content_type=asset.content_type,
            created_at=asset.created_at.isoformat(),
            get_url=get_url,
            thumb_url=thumb_url,
            duration=asset.duration,
        )


class AssetUpdateRequest(BaseModel):
    name: str | None
    parent_id: str | None
    status: str | None
    content_type: str | None
    duration: float | None = None


class AssetCreateRequest(BaseModel):
    workflow_id: str | None = None
    parent_id: str | None = None
    name: str
    content_type: str
    duration: float | None = None


class AssetList(BaseModel):
    next: str | None
    assets: list[Asset]


class TempAsset(BaseModel):
    get_url: str
    put_url: str



import warnings
import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import HWC3, resize_image_with_pad

class BinaryDetector:
    def __call__(self, input_image=None, bin_threshold=0, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        if "img" in kwargs:
            warnings.warn("img is deprecated, please use `input_image=...` instead.", DeprecationWarning)
            input_image = kwargs.pop("img")
        
        if input_image is None:
            raise ValueError("input_image must be defined.")

        if not isinstance(input_image, np.ndarray):
            input_image = np.array(input_image, dtype=np.uint8)
            output_type = output_type or "pil"
        else:
            output_type = output_type or "np"
        
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        img_gray = cv2.cvtColor(detected_map, cv2.COLOR_RGB2GRAY)
        if bin_threshold == 0 or bin_threshold == 255:
        # Otsu's threshold
            otsu_threshold, img_bin = cv2.threshold(img_gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
            print("Otsu threshold:", otsu_threshold)
        else:
            _, img_bin = cv2.threshold(img_gray, bin_threshold, 255, cv2.THRESH_BINARY_INV)
        
        detected_map = cv2.cvtColor(img_bin, cv2.COLOR_GRAY2RGB)
        detected_map = HWC3(remove_pad(255 - detected_map))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import numpy as np
import torch

from custom_manopth.manolayer import ManoLayer


def generate_random_hand(batch_size=1, ncomps=6, mano_root='mano/models'):
    nfull_comps = ncomps + 3  # Add global orientation dims to PCA
    random_pcapose = torch.rand(batch_size, nfull_comps)
    mano_layer = ManoLayer(mano_root=mano_root)
    verts, joints = mano_layer(random_pcapose)
    return {'verts': verts, 'joints': joints, 'faces': mano_layer.th_faces}


def display_hand(hand_info, mano_faces=None, ax=None, alpha=0.2, batch_idx=0, show=True):
    """
    Displays hand batch_idx in batch of hand_info, hand_info as returned by
    generate_random_hand
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    verts, joints = hand_info['verts'][batch_idx], hand_info['joints'][
        batch_idx]
    if mano_faces is None:
        ax.scatter(verts[:, 0], verts[:, 1], verts[:, 2], alpha=0.1)
    else:
        mesh = Poly3DCollection(verts[mano_faces], alpha=alpha)
        face_color = (141 / 255, 184 / 255, 226 / 255)
        edge_color = (50 / 255, 50 / 255, 50 / 255)
        mesh.set_edgecolor(edge_color)
        mesh.set_facecolor(face_color)
        ax.add_collection3d(mesh)
    ax.scatter(joints[:, 0], joints[:, 1], joints[:, 2], color='r')
    cam_equal_aspect_3d(ax, verts.numpy())
    if show:
        plt.show()


def cam_equal_aspect_3d(ax, verts, flip_x=False):
    """
    Centers view on cuboid containing hand and flips y and z axis
    and fixes azimuth
    """
    extents = np.stack([verts.min(0), verts.max(0)], axis=1)
    sz = extents[:, 1] - extents[:, 0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize / 2
    if flip_x:
        ax.set_xlim(centers[0] + r, centers[0] - r)
    else:
        ax.set_xlim(centers[0] - r, centers[0] + r)
    # Invert y and z axis
    ax.set_ylim(centers[1] + r, centers[1] - r)
    ax.set_zlim(centers[2] + r, centers[2] - r)



from .network import UNet
from .util import seg2img
import torch
import os
import cv2
from controlnet_aux.util import HWC3, resize_image_with_pad, common_input_validate, custom_hf_download, ANIFACESEG_MODEL_NAME
from huggingface_hub import hf_hub_download
from PIL import Image
from einops import rearrange
from .anime_segmentation import AnimeSegmentation
import numpy as np

class AnimeFaceSegmentor:
    def __init__(self, model, seg_model):
        self.model = model
        self.seg_model = seg_model

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=ANIFACESEG_MODEL_NAME, filename="UNet.pth", seg_filename="isnetis.ckpt"):
        model_path = custom_hf_download(pretrained_model_or_path, filename, subfolder="Annotators")
        seg_model_path = custom_hf_download("skytnt/anime-seg", seg_filename)

        model = UNet()
        ckpt = torch.load(model_path)
        model.load_state_dict(ckpt)
        model.eval()
        
        seg_model = AnimeSegmentation(seg_model_path)
        seg_model.net.eval()
        return cls(model, seg_model)

    def to(self, device):
        self.model.to(device)
        self.seg_model.net.to(device)
        return self

    def __call__(self, input_image, detect_resolution=512, output_type="pil", upscale_method="INTER_CUBIC", remove_background=True, **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        device = next(iter(self.model.parameters())).device

        with torch.no_grad():
            if remove_background:
                print(input_image.shape)
                mask, input_image = self.seg_model(input_image, 0) #Don't resize image as it is resized
            image_feed = torch.from_numpy(input_image).float().to(device)
            image_feed = rearrange(image_feed, 'h w c -> 1 c h w')
            image_feed = image_feed / 255
            seg = self.model(image_feed).squeeze(dim=0)
            result = seg2img(seg.cpu().detach().numpy())
        
        detected_map = HWC3(result)
        detected_map = remove_pad(detected_map)
        if remove_background:
            mask = remove_pad(mask)
            H, W, C = detected_map.shape
            tmp = np.zeros([H, W, C + 1])
            tmp[:,:,:C] = detected_map
            tmp[:,:,3:] = mask
            detected_map = tmp
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map[..., :3])
        
        return detected_map



# ------------------------------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------------------------------
# Modified from https://github.com/chengdazhi/Deformable-Convolution-V2-PyTorch/tree/pytorch_1.0.0
# ------------------------------------------------------------------------------------------------

# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from https://github.com/fundamentalvision/Deformable-DETR


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import torch
import torch.nn.functional as F
from torch.autograd import Function
from torch.autograd.function import once_differentiable

# if torch.cuda.is_available():
#     try:
#         import MultiScaleDeformableAttention as MSDA
#     except ModuleNotFoundError as e:
#         info_string = (
#             "\n\nPlease compile MultiScaleDeformableAttention CUDA op with the following commands:\n"
#             "\t`cd oneformer/modeling/pixel_decoder/ops`\n"
#             "\t`sh make.sh`\n"
#         )
#         raise ModuleNotFoundError(info_string)
# else:
#     MultiScaleDeformableAttention = None



class MSDeformAttnFunction(Function):
    @staticmethod
    def forward(ctx, value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, im2col_step):
        # ctx.im2col_step = im2col_step
        output = ms_deform_attn_core_pytorch(
            value, value_spatial_shapes, sampling_locations, attention_weights)
        # ctx.save_for_backward(value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights)
        return output

    # @staticmethod
    # @once_differentiable
    # def backward(ctx, grad_output):
    #     value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights = ctx.saved_tensors
    #     grad_value, grad_sampling_loc, grad_attn_weight = \
    #         MSDA.ms_deform_attn_backward(
    #             value, value_spatial_shapes, value_level_start_index, sampling_locations, attention_weights, grad_output, ctx.im2col_step)
    #
    #     return grad_value, None, None, grad_sampling_loc, grad_attn_weight, None


def ms_deform_attn_core_pytorch(value, value_spatial_shapes, sampling_locations, attention_weights):
    # for debug and test only,
    # need to use cuda version instead
    N_, S_, M_, D_ = value.shape
    _, Lq_, M_, L_, P_, _ = sampling_locations.shape
    value_list = value.split([H_ * W_ for H_, W_ in value_spatial_shapes], dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for lid_, (H_, W_) in enumerate(value_spatial_shapes):
        # N_, H_*W_, M_, D_ -> N_, H_*W_, M_*D_ -> N_, M_*D_, H_*W_ -> N_*M_, D_, H_, W_
        value_l_ = value_list[lid_].flatten(2).transpose(1, 2).reshape(N_*M_, D_, H_, W_)
        # N_, Lq_, M_, P_, 2 -> N_, M_, Lq_, P_, 2 -> N_*M_, Lq_, P_, 2
        sampling_grid_l_ = sampling_grids[:, :, :, lid_].transpose(1, 2).flatten(0, 1)
        # N_*M_, D_, Lq_, P_
        sampling_value_l_ = F.grid_sample(value_l_, sampling_grid_l_,
                                          mode='bilinear', padding_mode='zeros', align_corners=False)
        sampling_value_list.append(sampling_value_l_)
    # (N_, Lq_, M_, L_, P_) -> (N_, M_, Lq_, L_, P_) -> (N_, M_, 1, Lq_, L_*P_)
    attention_weights = attention_weights.transpose(1, 2).reshape(N_*M_, 1, Lq_, L_*P_)
    output = (torch.stack(sampling_value_list, dim=-2).flatten(-2) * attention_weights).sum(-1).view(N_, M_*D_, Lq_)
    return output.transpose(1, 2).contiguous()



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management
import numpy as np
import warnings
from controlnet_aux.dwpose import DwposeDetector, AnimalposeDetector
import os
import json

DWPOSE_MODEL_NAME = "yzd-v/DWPose"
#Trigger startup caching for onnxruntime
GPU_PROVIDERS = ["CUDAExecutionProvider", "DirectMLExecutionProvider", "OpenVINOExecutionProvider", "ROCMExecutionProvider", "CoreMLExecutionProvider"]
def check_ort_gpu():
    try:
        import onnxruntime as ort
        for provider in GPU_PROVIDERS:
            if provider in ort.get_available_providers():
                return True
        return False
    except:
        return False

if not os.environ.get("DWPOSE_ONNXRT_CHECKED"):
    if check_ort_gpu():
        print("DWPose: Onnxruntime with acceleration providers detected")
    else:
        warnings.warn("DWPose: Onnxruntime not found or doesn't come with acceleration providers, switch to OpenCV with CPU device. DWPose might run very slowly")
        os.environ['AUX_ORT_PROVIDERS'] = ''
    os.environ["DWPOSE_ONNXRT_CHECKED"] = '1'

class DWPose_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        input_types = create_node_input_types(
            detect_hand=(["enable", "disable"], {"default": "enable"}),
            detect_body=(["enable", "disable"], {"default": "enable"}),
            detect_face=(["enable", "disable"], {"default": "enable"})
        )
        input_types["optional"] = {
            **input_types["optional"],
            "bbox_detector": (
                ["yolox_l.torchscript.pt", "yolox_l.onnx", "yolo_nas_l_fp16.onnx", "yolo_nas_m_fp16.onnx", "yolo_nas_s_fp16.onnx"],
                {"default": "yolox_l.torchscript.pt"}
            ),
            "pose_estimator": (["dw-ll_ucoco_384_bs5.torchscript.pt", "dw-ll_ucoco_384.onnx", "dw-ll_ucoco.onnx"], {"default": "dw-ll_ucoco_384_bs5.torchscript.pt"})
        }
        return input_types

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "estimate_pose"

    CATEGORY = "ControlNet Preprocessors/Faces and Poses Estimators"

    def estimate_pose(self, image, detect_hand, detect_body, detect_face, resolution=512, bbox_detector="yolox_l.onnx", pose_estimator="dw-ll_ucoco_384.onnx", **kwargs):
        if bbox_detector == "yolox_l.onnx":
            yolo_repo = DWPOSE_MODEL_NAME
        elif "yolox" in bbox_detector:
            yolo_repo = "hr16/yolox-onnx"
        elif "yolo_nas" in bbox_detector:
            yolo_repo = "hr16/yolo-nas-fp16"
        else:
            raise NotImplementedError(f"Download mechanism for {bbox_detector}")

        if pose_estimator == "dw-ll_ucoco_384.onnx":
            pose_repo = DWPOSE_MODEL_NAME
        elif pose_estimator.endswith(".onnx"):
            pose_repo = "hr16/UnJIT-DWPose"
        elif pose_estimator.endswith(".torchscript.pt"):
            pose_repo = "hr16/DWPose-TorchScript-BatchSize5"
        else:
            raise NotImplementedError(f"Download mechanism for {pose_estimator}")

        model = DwposeDetector.from_pretrained(
            pose_repo,
            yolo_repo,
            det_filename=bbox_detector, pose_filename=pose_estimator,
            torchscript_device=model_management.get_torch_device()
        )
        detect_hand = detect_hand == "enable"
        detect_body = detect_body == "enable"
        detect_face = detect_face == "enable"
        self.openpose_dicts = []
        def func(image, **kwargs):
            pose_img, openpose_dict = model(image, **kwargs)
            self.openpose_dicts.append(openpose_dict)
            return pose_img

        out = common_annotator_call(func, image, include_hand=detect_hand, include_face=detect_face, include_body=detect_body, image_and_json=True, resolution=resolution)
        del model
        return {
            'ui': { "openpose_json": [json.dumps(self.openpose_dicts, indent=4)] },
            "result": (out, )
        }

class AnimalPose_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            bbox_detector = (
                ["yolox_l.torchscript.pt", "yolox_l.onnx", "yolo_nas_l_fp16.onnx", "yolo_nas_m_fp16.onnx", "yolo_nas_s_fp16.onnx"],
                {"default": "yolox_l.torchscript.pt"}
            ),
            pose_estimator = (["rtmpose-m_ap10k_256_bs5.torchscript.pt", "rtmpose-m_ap10k_256.onnx"], {"default": "rtmpose-m_ap10k_256_bs5.torchscript.pt"})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "estimate_pose"

    CATEGORY = "ControlNet Preprocessors/Faces and Poses Estimators"

    def estimate_pose(self, image, resolution=512, bbox_detector="yolox_l.onnx", pose_estimator="rtmpose-m_ap10k_256.onnx", **kwargs):
        if bbox_detector == "yolox_l.onnx":
            yolo_repo = DWPOSE_MODEL_NAME
        elif "yolox" in bbox_detector:
            yolo_repo = "hr16/yolox-onnx"
        elif "yolo_nas" in bbox_detector:
            yolo_repo = "hr16/yolo-nas-fp16"
        else:
            raise NotImplementedError(f"Download mechanism for {bbox_detector}")

        if pose_estimator == "dw-ll_ucoco_384.onnx":
            pose_repo = DWPOSE_MODEL_NAME
        elif pose_estimator.endswith(".onnx"):
            pose_repo = "hr16/UnJIT-DWPose"
        elif pose_estimator.endswith(".torchscript.pt"):
            pose_repo = "hr16/DWPose-TorchScript-BatchSize5"
        else:
            raise NotImplementedError(f"Download mechanism for {pose_estimator}")

        model = AnimalposeDetector.from_pretrained(
            pose_repo,
            yolo_repo,
            det_filename=bbox_detector, pose_filename=pose_estimator,
            torchscript_device=model_management.get_torch_device()
        )

        self.openpose_dicts = []
        def func(image, **kwargs):
            pose_img, openpose_dict = model(image, **kwargs)
            self.openpose_dicts.append(openpose_dict)
            return pose_img

        out = common_annotator_call(func, image, image_and_json=True, resolution=resolution)
        del model
        return {
            'ui': { "openpose_json": [json.dumps(self.openpose_dicts, indent=4)] },
            "result": (out, )
        }

NODE_CLASS_MAPPINGS = {
    "DWPreprocessor": DWPose_Preprocessor,
    "AnimalPosePreprocessor": AnimalPose_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "DWPreprocessor": "DWPose Estimator",
    "AnimalPosePreprocessor": "AnimalPose Estimator (AP10K)"
}


import torch.nn as nn
import numpy as np

#torch.set_printoptions(precision=10)


class _bn_relu_conv(nn.Module):
    def __init__(self, in_filters, nb_filters, fw, fh, subsample=1):
        super(_bn_relu_conv, self).__init__()
        self.model = nn.Sequential(
            nn.BatchNorm2d(in_filters, eps=1e-3),
            nn.LeakyReLU(0.2),
            nn.Conv2d(in_filters, nb_filters, (fw, fh), stride=subsample, padding=(fw//2, fh//2), padding_mode='zeros')
        )

    def forward(self, x):
        return self.model(x)

        # the following are for debugs
        print("****", np.max(x.cpu().numpy()), np.min(x.cpu().numpy()), np.mean(x.cpu().numpy()), np.std(x.cpu().numpy()), x.shape)
        for i,layer in enumerate(self.model):
            if i != 2:
                x = layer(x)
            else:
                x = layer(x)
                #x = nn.functional.pad(x, (1, 1, 1, 1), mode='constant', value=0)
            print("____", np.max(x.cpu().numpy()), np.min(x.cpu().numpy()), np.mean(x.cpu().numpy()), np.std(x.cpu().numpy()), x.shape)
            print(x[0])
        return x


class _u_bn_relu_conv(nn.Module):
    def __init__(self, in_filters, nb_filters, fw, fh, subsample=1):
        super(_u_bn_relu_conv, self).__init__()
        self.model = nn.Sequential(
            nn.BatchNorm2d(in_filters, eps=1e-3),
            nn.LeakyReLU(0.2),
            nn.Conv2d(in_filters, nb_filters, (fw, fh), stride=subsample, padding=(fw//2, fh//2)),
            nn.Upsample(scale_factor=2, mode='nearest')
        )

    def forward(self, x):
        return self.model(x)



class _shortcut(nn.Module):
    def __init__(self, in_filters, nb_filters, subsample=1):
        super(_shortcut, self).__init__()
        self.process = False
        self.model = None
        if in_filters != nb_filters or subsample != 1:
            self.process = True
            self.model = nn.Sequential(
                    nn.Conv2d(in_filters, nb_filters, (1, 1), stride=subsample)
                )

    def forward(self, x, y):
        #print(x.size(), y.size(), self.process)
        if self.process:
            y0 = self.model(x)
            #print("merge+", torch.max(y0+y), torch.min(y0+y),torch.mean(y0+y), torch.std(y0+y), y0.shape)
            return y0 + y
        else:
            #print("merge", torch.max(x+y), torch.min(x+y),torch.mean(x+y), torch.std(x+y), y.shape)
            return x + y

class _u_shortcut(nn.Module):
    def __init__(self, in_filters, nb_filters, subsample):
        super(_u_shortcut, self).__init__()
        self.process = False
        self.model = None
        if in_filters != nb_filters:
            self.process = True
            self.model = nn.Sequential(
                nn.Conv2d(in_filters, nb_filters, (1, 1), stride=subsample, padding_mode='zeros'),
                nn.Upsample(scale_factor=2, mode='nearest')
            )

    def forward(self, x, y):
        if self.process:
            return self.model(x) + y
        else:
            return x + y


class basic_block(nn.Module):
    def __init__(self, in_filters, nb_filters, init_subsample=1):
        super(basic_block, self).__init__()
        self.conv1 = _bn_relu_conv(in_filters, nb_filters, 3, 3, subsample=init_subsample)
        self.residual = _bn_relu_conv(nb_filters, nb_filters, 3, 3)
        self.shortcut = _shortcut(in_filters, nb_filters, subsample=init_subsample)

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.residual(x1)
        return self.shortcut(x, x2)

class _u_basic_block(nn.Module):
    def __init__(self, in_filters, nb_filters, init_subsample=1):
        super(_u_basic_block, self).__init__()
        self.conv1 = _u_bn_relu_conv(in_filters, nb_filters, 3, 3, subsample=init_subsample)
        self.residual = _bn_relu_conv(nb_filters, nb_filters, 3, 3)
        self.shortcut = _u_shortcut(in_filters, nb_filters, subsample=init_subsample)

    def forward(self, x):
        y = self.residual(self.conv1(x))
        return self.shortcut(x, y)


class _residual_block(nn.Module):
    def __init__(self, in_filters, nb_filters, repetitions, is_first_layer=False):
        super(_residual_block, self).__init__()
        layers = []
        for i in range(repetitions):
            init_subsample = 1
            if i == repetitions - 1 and not is_first_layer:
                init_subsample = 2
            if i == 0:
                l = basic_block(in_filters=in_filters, nb_filters=nb_filters, init_subsample=init_subsample)
            else:
                l = basic_block(in_filters=nb_filters, nb_filters=nb_filters, init_subsample=init_subsample)
            layers.append(l)

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class _upsampling_residual_block(nn.Module):
    def __init__(self, in_filters, nb_filters, repetitions):
        super(_upsampling_residual_block, self).__init__()
        layers = []
        for i in range(repetitions):
            l = None
            if i == 0: 
                l = _u_basic_block(in_filters=in_filters, nb_filters=nb_filters)#(input)
            else:
                l = basic_block(in_filters=nb_filters, nb_filters=nb_filters)#(input)
            layers.append(l)

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class res_skip(nn.Module):

    def __init__(self):
        super(res_skip, self).__init__()
        self.block0 = _residual_block(in_filters=1, nb_filters=24, repetitions=2, is_first_layer=True)#(input)
        self.block1 = _residual_block(in_filters=24, nb_filters=48, repetitions=3)#(block0)
        self.block2 = _residual_block(in_filters=48, nb_filters=96, repetitions=5)#(block1)
        self.block3 = _residual_block(in_filters=96, nb_filters=192, repetitions=7)#(block2)
        self.block4 = _residual_block(in_filters=192, nb_filters=384, repetitions=12)#(block3)
        
        self.block5 = _upsampling_residual_block(in_filters=384, nb_filters=192, repetitions=7)#(block4)
        self.res1 = _shortcut(in_filters=192, nb_filters=192)#(block3, block5, subsample=(1,1))

        self.block6 = _upsampling_residual_block(in_filters=192, nb_filters=96, repetitions=5)#(res1)
        self.res2 = _shortcut(in_filters=96, nb_filters=96)#(block2, block6, subsample=(1,1))

        self.block7 = _upsampling_residual_block(in_filters=96, nb_filters=48, repetitions=3)#(res2)
        self.res3 = _shortcut(in_filters=48, nb_filters=48)#(block1, block7, subsample=(1,1))

        self.block8 = _upsampling_residual_block(in_filters=48, nb_filters=24, repetitions=2)#(res3)
        self.res4 = _shortcut(in_filters=24, nb_filters=24)#(block0,block8, subsample=(1,1))

        self.block9 = _residual_block(in_filters=24, nb_filters=16, repetitions=2, is_first_layer=True)#(res4)
        self.conv15 = _bn_relu_conv(in_filters=16, nb_filters=1, fh=1, fw=1, subsample=1)#(block7)

    def forward(self, x):
        x0 = self.block0(x)
        x1 = self.block1(x0)
        x2 = self.block2(x1)
        x3 = self.block3(x2)
        x4 = self.block4(x3)

        x5 = self.block5(x4)
        res1 = self.res1(x3, x5)

        x6 = self.block6(res1)
        res2 = self.res2(x2, x6)

        x7 = self.block7(res2)
        res3 = self.res3(x1, x7)

        x8 = self.block8(res3)
        res4 = self.res4(x0, x8)

        x9 = self.block9(res4)
        y = self.conv15(x9)

        return y


import io
import numpy as np
import pydub
from pydub import AudioSegment
import pydub.effects
from typing import List, Tuple, cast
from pydub.silence import detect_nonsilent
from pydub.silence import split_on_silence


def convert_to_float(audio_data: np.ndarray):
    """
    Converts audio data from integer format to float format.

    Args:
        audio_data (np.ndarray): Audio data in integer format.

    Returns:
        np.ndarray: Audio data in float format.
    """
    dtype = audio_data.dtype

    if dtype == np.float32 or dtype == np.float64:
        return audio_data
    if dtype == np.int16:
        return audio_data / 2**15
    if dtype == np.int32:
        return audio_data / 2**31
    if dtype == np.int64:
        return audio_data / 2**63

    raise Exception(f"Unsupported dtype: {dtype}")


def scale_audios_to_shortest_duration(audios: List[AudioSegment]):
    """
    Scales a list of audios to the shortest duration, meaning that
    all audios will have the same duration.

    Args:
        audios (List[AudioSegment]): A list of audios to scale.

    Returns:
        List[AudioSegment]: A list of scaled audio segments.
    """
    # Find the shortest duration
    min_duration = min([len(audio) for audio in audios])

    # Scale audios to the shortest duration
    scaled_audios = [audio[:min_duration] for audio in audios]

    return scaled_audios


def resize_audio(audio: AudioSegment, duration: float) -> AudioSegment:
    """
    Resize the audio to the given duration.

    Args:
        audio (AudioSegment): The audio to resize.
        duration (float): The duration to resize the audio to, in milliseconds.

    Returns:
        AudioSegment: The resized audio segment.
    """
    if len(audio) > duration:
        return cast(AudioSegment, audio[:duration])
    elif len(audio) < duration:
        padding = AudioSegment.silent(duration=int(len(audio) - duration))
        return audio + padding
    else:
        return audio


def concatenate_audios(audios: List[AudioSegment]) -> AudioSegment:
    """
    Concatenate a list of audio segments

    Args:
        audios (List[AudioSegment]): A list of audio segments to concatenate.

    Returns:
        AudioSegment: The concatenated audio segment.
    """
    concatenated_audio = AudioSegment.empty()
    for audio in audios:
        concatenated_audio += audio
    return concatenated_audio


def remove_silence(audio: AudioSegment) -> AudioSegment:
    """
    Remove silence from an audio segment.
    This is done by detecting nonsilent ranges and concatenating them.

    Args:
        audio (AudioSegment): The audio to remove silence from.

    Returns:
        AudioSegment: The audio with silence removed.
    """
    nonsilent_ranges = detect_nonsilent(audio, min_silence_len=100, silence_thresh=-32)
    return concatenate_audios(
        [cast(AudioSegment, audio[start:end]) for start, end in nonsilent_ranges]
    )


def segment_audio(audio: AudioSegment) -> List[AudioSegment]:
    """
    Segment an audio into multiple audio segments.

    Args:
        audio (AudioSegment): The audio to segment.

    Returns:
        List[AudioSegment]: A list of segmented audio segments.
    """
    res = split_on_silence(
        audio, min_silence_len=500, silence_thresh=-32, keep_silence=100
    )
    return [cast(AudioSegment, segment) for segment in res]


def resample_audio(audio: AudioSegment, sample_rate: int) -> AudioSegment:
    """
    Resample the audio to the given sample rate.

    Args:
        audio (AudioSegment): The audio to resample.
        sample_rate (int): The sample rate to resample the audio to.

    Returns:
        AudioSegment: The resampled audio segment.
    """
    return audio.set_frame_rate(sample_rate)


def get_audio_metadata(audio: AudioSegment) -> dict:
    """
    Get metadata for an audio segment.

    Args:
        audio (AudioSegment): The audio to get metadata for.

    Returns:
        dict: A dictionary containing the audio metadata.
    """
    return {
        "channels": audio.channels,
        "frame_rate": audio.frame_rate,
        "frame_width": audio.frame_width,
        "length": len(audio),
    }


def convert_audio_format(audio: AudioSegment, format: str) -> AudioSegment:
    """
    Convert the audio to the given format.

    Args:
        audio (AudioSegment): The audio to convert.
        format (str): The format to convert the audio to.

    Returns:
        AudioSegment: The converted audio segment.
    """
    buffer = io.BytesIO()
    audio.export(buffer, format=format)
    buffer.seek(0)
    return AudioSegment.from_file(buffer, format=format)


def export_audio(
    audio: AudioSegment, file_path: str, format: str, codec: str, bitrate: str
) -> None:
    """
    Export the audio to a file.

    Args:
        audio (AudioSegment): The audio to export.
        file_path (str): The path to the file to export the audio to.
        format (str): The format to export the audio in.
        codec (str): The codec to use for the exported audio.
        bitrate (str): The bitrate to use for the exported audio.
    """
    audio.export(file_path, format=format, codec=codec, bitrate=bitrate)


def convert_audio_to_bytes(audio: AudioSegment) -> bytes:
    """
    Convert the audio to bytes.

    Args:
        audio (AudioSegment): The audio to convert.

    Returns:
        bytes: The audio as bytes.
    """
    with io.BytesIO() as buffer:
        audio.export(buffer, format="MP3")
        return buffer.getvalue()


def normalize_audio(audio: AudioSegment) -> AudioSegment:
    """
    Normalize the audio.

    Args:
        audio (AudioSegment): The audio to normalize.

    Returns:
        AudioSegment: The normalized audio segment.
    """
    return pydub.effects.normalize(audio)


def numpy_to_audio_segment(arr: np.ndarray, sample_rate=44100) -> AudioSegment:
    """
    Convert a numpy array to an audio segment.

    Args:
        arr (np.ndarray): The numpy array to convert.
        sample_rate (int): The sample rate of the audio segment.

    Returns:
        AudioSegment: The audio segment.
    """
    # Convert the float array to int16 format, which is used by WAV files.
    arr_int16 = np.int16(arr * 32767.0).tobytes()

    # Create a pydub AudioSegment from raw data.
    return AudioSegment(arr_int16, sample_width=2, frame_rate=sample_rate, channels=1)



import warnings
from typing import Union

import cv2
import numpy as np
from PIL import Image

from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad
from .mediapipe_face_common import generate_annotation


class MediapipeFaceDetector:
    def __call__(self,
                 input_image: Union[np.ndarray, Image.Image] = None,
                 max_faces: int = 1,
                 min_confidence: float = 0.5,
                 output_type: str = "pil",
                 detect_resolution: int = 512,
                 image_resolution: int = 512,
                 upscale_method="INTER_CUBIC",
                 **kwargs):

        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        detected_map = generate_annotation(detected_map, max_faces, min_confidence)
        detected_map = remove_pad(HWC3(detected_map))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map



"""
This file contains a Processor that can be used to process images with controlnet aux processors
"""
import io
import logging
from typing import Dict, Optional, Union

from PIL import Image

from controlnet_aux import (CannyDetector, ContentShuffleDetector, HEDdetector,
                            LeresDetector, LineartAnimeDetector,
                            LineartDetector, MediapipeFaceDetector,
                            MidasDetector, MLSDdetector, NormalBaeDetector,
                            OpenposeDetector, PidiNetDetector, ZoeDetector, TileDetector)

LOGGER = logging.getLogger(__name__)


MODELS = {
    # checkpoint models
    'scribble_hed': {'class': HEDdetector, 'checkpoint': True},
    'softedge_hed': {'class': HEDdetector, 'checkpoint': True},
    'scribble_hedsafe': {'class': HEDdetector, 'checkpoint': True},
    'softedge_hedsafe': {'class': HEDdetector, 'checkpoint': True},
    'depth_midas': {'class': MidasDetector, 'checkpoint': True},
    'mlsd': {'class': MLSDdetector, 'checkpoint': True},
    'openpose': {'class': OpenposeDetector, 'checkpoint': True},
    'openpose_face': {'class': OpenposeDetector, 'checkpoint': True},
    'openpose_faceonly': {'class': OpenposeDetector, 'checkpoint': True},
    'openpose_full': {'class': OpenposeDetector, 'checkpoint': True},
    'openpose_hand': {'class': OpenposeDetector, 'checkpoint': True},
    'scribble_pidinet': {'class': PidiNetDetector, 'checkpoint': True},
    'softedge_pidinet': {'class': PidiNetDetector, 'checkpoint': True},
    'scribble_pidsafe': {'class': PidiNetDetector, 'checkpoint': True},
    'softedge_pidsafe': {'class': PidiNetDetector, 'checkpoint': True},
    'normal_bae': {'class': NormalBaeDetector, 'checkpoint': True},
    'lineart_coarse': {'class': LineartDetector, 'checkpoint': True},
    'lineart_realistic': {'class': LineartDetector, 'checkpoint': True},
    'lineart_anime': {'class': LineartAnimeDetector, 'checkpoint': True},
    'depth_zoe': {'class': ZoeDetector, 'checkpoint': True}, 
    'depth_leres': {'class': LeresDetector, 'checkpoint': True}, 
    'depth_leres++': {'class': LeresDetector, 'checkpoint': True}, 
    # instantiate
    'shuffle': {'class': ContentShuffleDetector, 'checkpoint': False},
    'mediapipe_face': {'class': MediapipeFaceDetector, 'checkpoint': False},
    'canny': {'class': CannyDetector, 'checkpoint': False},
    'tile': {'class': TileDetector, 'checkpoint': False},
}


MODEL_PARAMS = {
    'scribble_hed': {'scribble': True},
    'softedge_hed': {'scribble': False},
    'scribble_hedsafe': {'scribble': True, 'safe': True},
    'softedge_hedsafe': {'scribble': False, 'safe': True},
    'depth_midas': {},
    'mlsd': {},
    'openpose': {'include_body': True, 'include_hand': False, 'include_face': False},
    'openpose_face': {'include_body': True, 'include_hand': False, 'include_face': True},
    'openpose_faceonly': {'include_body': False, 'include_hand': False, 'include_face': True},
    'openpose_full': {'include_body': True, 'include_hand': True, 'include_face': True},
    'openpose_hand': {'include_body': False, 'include_hand': True, 'include_face': False},
    'scribble_pidinet': {'safe': False, 'scribble': True},
    'softedge_pidinet': {'safe': False, 'scribble': False},
    'scribble_pidsafe': {'safe': True, 'scribble': True},
    'softedge_pidsafe': {'safe': True, 'scribble': False},
    'normal_bae': {},
    'lineart_realistic': {'coarse': False},
    'lineart_coarse': {'coarse': True},
    'lineart_anime': {},
    'canny': {},
    'shuffle': {},
    'depth_zoe': {},
    'depth_leres': {'boost': False},
    'depth_leres++': {'boost': True},
    'mediapipe_face': {},
    'tile': {},
}

CHOICES = f"Choices for the processor are {list(MODELS.keys())}"


class Processor:
    def __init__(self, processor_id: str, params: Optional[Dict] = None) -> None:
        """Processor that can be used to process images with controlnet aux processors

        Args:
            processor_id (str): processor name, options are 'hed, midas, mlsd, openpose,
                                pidinet, normalbae, lineart, lineart_coarse, lineart_anime,
                                canny, content_shuffle, zoe, mediapipe_face, tile'
            params (Optional[Dict]): parameters for the processor
        """
        LOGGER.info("Loading %s".format(processor_id))

        if processor_id not in MODELS:
            raise ValueError(f"{processor_id} is not a valid processor id. Please make sure to choose one of {', '.join(MODELS.keys())}")

        self.processor_id = processor_id
        self.processor = self.load_processor(self.processor_id)

        # load default params
        self.params = MODEL_PARAMS[self.processor_id]
        # update with user params
        if params:
            self.params.update(params)

    def load_processor(self, processor_id: str) -> 'Processor':
        """Load controlnet aux processors

        Args:
            processor_id (str): processor name

        Returns:
            Processor: controlnet aux processor
        """
        processor = MODELS[processor_id]['class']

        # check if the proecssor is a checkpoint model
        if MODELS[processor_id]['checkpoint']:
            processor = processor.from_pretrained("lllyasviel/Annotators")
        else:
            processor = processor()
        return processor

    def __call__(self, image: Union[Image.Image, bytes],
                 to_pil: bool = True) -> Union[Image.Image, bytes]:
        """processes an image with a controlnet aux processor

        Args:
            image (Union[Image.Image, bytes]): input image in bytes or PIL Image
            to_pil (bool): whether to return bytes or PIL Image

        Returns:
            Union[Image.Image, bytes]: processed image in bytes or PIL Image
        """
        # check if bytes or PIL Image
        if isinstance(image, bytes):
            image = Image.open(io.BytesIO(image)).convert("RGB")

        processed_image = self.processor(image, **self.params)

        if to_pil:
            return processed_image
        else:
            output_bytes = io.BytesIO()
            processed_image.save(output_bytes, format='JPEG')
            return output_bytes.getvalue()



import warnings
import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import get_upscale_method, common_input_validate, HWC3


class TileDetector:
    def __call__(self, input_image=None, pyrUp_iters=3, output_type=None, upscale_method="INTER_AREA", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        H, W, _ = input_image.shape
        H = int(np.round(H / 64.0)) * 64
        W = int(np.round(W / 64.0)) * 64
        detected_map = cv2.resize(input_image, (W // (2 ** pyrUp_iters), H // (2 ** pyrUp_iters)),
                                  interpolation=get_upscale_method(upscale_method))
        detected_map = HWC3(detected_map)

        for _ in range(pyrUp_iters):
            detected_map = cv2.pyrUp(detected_map)

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)

        return detected_map
